import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import cg
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.widgets import Slider

try:
    from Element_Stiffness import Stiffness_Matrix
    from Meshing import MeshingClass
except:
    from KIB_LAP.Scheibe.Element_Stiffness import Stiffness_Matrix
    from KIB_LAP.Scheibe.Meshing import MeshingClass
    
import pandas as pd
from tabulate import *


class Assembled_Matrices:
    def __init__(
        self,
        _mesh,
        EL,
        NL,
        EL_Matrices,
        _ElementLoadVector,
        _ElementLoadVectorNodes,
        _ElementLoadVectorAreaLoads,
        _ElementLoadVectorAreaLoadsNodes,
        load_input="csv",
        bc_input="csv",
    ):
        # Meshing
        self.mesh = _mesh

        self.EL = EL
        self.NL = NL
        # Generate element stiffness matrices
        self.ElementLoadVector = _ElementLoadVector
        self.ElementLoadVectorNodes = _ElementLoadVectorNodes
        self.ElementLoadVectorAreaLoads = _ElementLoadVectorAreaLoads
        self.ElementLoadVectorAreaLoadsNodes = _ElementLoadVectorAreaLoadsNodes
        #  Assemble Stiffness matrices
        self._EL = np.array(EL)
        self._NoE = self._EL.shape[0]
        self._NPE = self._EL.shape[1]
        self.NoN = len(NL)
        self.PD = len(NL[0])
        self._EL_Matrices = np.array(EL_Matrices)
        self.load_input = load_input
        self.bc_input = bc_input

        self.K_Assemble = np.zeros((self.NoN * self.PD, self.NoN * self.PD))
        self.K_Assemble_BC = None
        self.Index_Vector_BC = []

    def assemble_K(self):
        for i in range(self._NoE):
            Index_Vector = np.zeros(self._NPE * 2, dtype=int)
            for j in range(self._NPE):
                Index_Vector[j] = int(self._EL[i, j] * 2 - 2)
                Index_Vector[j + self._NPE] = int(self._EL[i, j] * 2 - 1)

            for row in range(len(Index_Vector)):
                ind_row = Index_Vector[row]
                for col in range(len(Index_Vector)):
                    ind_col = Index_Vector[col]
                    self.K_Assemble[ind_row, ind_col] += self._EL_Matrices[i][row][col]

    def Load_BC(self):
        if self.bc_input == "console":
            # Implementieren Sie hier die Konsoleneingabe
            pass
        else:
            try:
                self.BC = pd.DataFrame(
                    pd.read_csv("Boundary_Conditions/Spring_Elements.csv")
                )
            except:
                """
                This exception is used, if the main file is started from a subfolder (e.g. the parametric studies or testing) \n
                folder \n
                """
                self.BC = pd.DataFrame(
                    pd.read_csv("../Boundary_Conditions/Spring_Elements.csv")
                )

    def apply_BC(self):
        self.K_Assemble_BC = np.copy(self.K_Assemble)
        self.Index_Vector_BC = np.zeros(0, dtype=int)
        self.stiffness_spring = []

        for i in range(len(self.BC["No"])):
            node = self.BC["No"][i]
            dof = self.BC["DOF"][i]
            stiffness = self.BC["cf in [MN/m]"][i]
            print("Stiffness ", stiffness)
            print(node)
            # Überprüfen, ob der Knoten ein numerischer Wert ist
            if str(node).isdigit():
                node = int(node)
                if dof == "x":
                    self.Index_Vector_BC = np.append(self.Index_Vector_BC, (node - 1) * 2)
                elif dof == "z":
                    self.Index_Vector_BC = np.append(self.Index_Vector_BC, (node - 1) * 2 + 1)
            elif node == "left":
                nodes = self.mesh.get_left_border_nodes()
                print("LEFT DOF")
                print(dof)
                for p in nodes:
                    if dof == "x":
                        self.Index_Vector_BC = np.append(self.Index_Vector_BC, (p - 1) * 2)
                    elif dof == "z":
                        self.Index_Vector_BC = np.append(self.Index_Vector_BC, (p - 1) * 2 + 1)
                    else:
                        print("Exception. No valid coordinate input for the left boundary.")
            elif node == "right":
                nodes = self.mesh.get_right_border_nodes()
                for p in nodes:
                    if dof == "x":
                        self.Index_Vector_BC = np.append(self.Index_Vector_BC, (p - 1) * 2)
                    elif dof == "z":
                        self.Index_Vector_BC = np.append(self.Index_Vector_BC, (p - 1) * 2 + 1)
                    else:
                        print("Exception. No valid coordinate input for the right boundary.")
            else:
                print("The input isn't usable! Please change the DOF of the boundary condition at Position "
                    + str(i + 1) + "\n")
            
            try:
                self.stiffness_spring.append(stiffness)
            except:
                print("Error with the spring stiffness")

        self.Index_Vector_BC = sorted(self.Index_Vector_BC)

        p = 0
        for index in sorted(self.Index_Vector_BC, reverse=True):
            self.K_Assemble_BC[index][index] += self.stiffness_spring[p]
            p+=1

        print(self.Index_Vector_BC)

    def LoadInput(self):
        self.Einzellasten = False
        self.Linienlasten = False
        if self.load_input == "console":
            # Implementieren Sie hier die Konsoleneingabe
            pass
        else:
            try:
                self.NodalForces = pd.DataFrame(pd.read_csv("Loading/Nodal_Loads.csv"))
                self.ElementLoads = pd.DataFrame(
                    pd.read_csv("Loading/Element_Loads.csv")
                )
            except:
                self.NodalForces = pd.DataFrame(
                    pd.read_csv("../Loading/Nodal_Loads.csv")
                )
                self.ElementLoads = pd.DataFrame(
                    pd.read_csv("../Loading/Element_Loads.csv")
                )

    def GenerateLoadVector(self):
        self.Load_Vector = np.zeros(len(self.K_Assemble_BC))

        for i in range(0, len(self.NodalForces["DOF"])):
            if self.NodalForces["DOF"][i] == "x":
                index_load = (self.NodalForces["No"][i] - 1) * 2  # x-Direction
                self.Load_Vector[index_load] = self.NodalForces["F in [MN]"][i]
            elif self.NodalForces["DOF"][i] == "z":
                index_load = (self.NodalForces["No"][i] - 1) * 2 + 1  # z-Direction
                self.Load_Vector[index_load] = self.NodalForces["F in [MN]"][i]

            else:
                print(
                    "The input isn't usable! Please change the DOF of the boundary condition at Position "
                    + i
                    + 1
                    + "\n"
                )

        for i in range(0, len(self.ElementLoadVectorNodes)):
            index_load_x = int((self.ElementLoadVectorNodes[i] - 1) * 2)
            index_load_z = int((self.ElementLoadVectorNodes[i] - 1) * 2 + 1)

            self.Load_Vector[index_load_x] += self.ElementLoadVector[0][i]
            self.Load_Vector[index_load_z] += self.ElementLoadVector[1][i]

        for i in range(0, len(self.ElementLoadVectorAreaLoadsNodes), 1):
            p = 0
            for j in range(0, len(self.ElementLoadVectorAreaLoadsNodes[i]), 1):
                index_load_x = int((self.ElementLoadVectorAreaLoadsNodes[i][j] - 1) * 2)
                index_load_z = int(
                    (self.ElementLoadVectorAreaLoadsNodes[i][j] - 1) * 2 + 1
                )

                self.Load_Vector[index_load_x] += self.ElementLoadVectorAreaLoads[
                    p * 2
                ][i]
                self.Load_Vector[index_load_z] += self.ElementLoadVectorAreaLoads[
                    p * 2 + 1
                ][i]

                p += 1

    def get_assembled(self):
        return self.K_Assemble

    def get_assembled_BC(self):
        return self.K_Assemble_BC

    def Solve(self):
        if len(self.K_Assemble) < 20:
            self.x_reduced = np.linalg.solve(self.K_Assemble_BC, self.Load_Vector)
        else:
            self.x_reduced, sparse_info = cg(self.K_Assemble_BC, self.Load_Vector)
            print("Information about sparse solver ")
            sparse_info

        self.x_disp = self.x_reduced[::2]
        self.z_disp = self.x_reduced[1::2]

    def StoreElementDisplacements(self):
        self.disp_element_matrix = np.zeros((8, self._NoE))
        for row in range(0, len(self.EL), 1):
            p = 0
            for col in range(0, len(self.EL[0]), 1):
                index_x = int((self.EL[row][col] - 1) * 2)
                index_z = int((self.EL[row][col] - 1) * 2 + 1)

                disp_x = self.x_reduced[index_x]
                disp_z = self.x_reduced[index_z]

                self.disp_element_matrix[p * 2][row] = disp_x
                self.disp_element_matrix[p * 2 + 1][row] = disp_z

                p += 1


    def __str__(self):
        output = "Assembled stiffness matrix\n"
        output += tabulate(self.K_Assemble) + "\n"
        output += "Assembled stiffness matrix with boundary conditions\n"
        output += tabulate(self.K_Assemble_BC)
        return output



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from Kontinuum_Schwingung import *

class DuhamelSolver:
    def __init__(self, k, m, zeta, dt, duration, loading = "Pedestrian", parameters = [2,1.3,800,60], m2_imp = 0):
        """
        loading = Pedestrian, Harmonic, Impulse
        parameters = [Frequency in Hz, 
        Speed in [m/s],
        First harmonic Loading in [N],  
        Length of the Bridge in [m]]
        """
        self.k = k
        self.m = m + m2_imp      # mass of the system, including the impact mass
        self.m2 = m2_imp        # impact mass, if impact response is calculated
        self.zeta = zeta
        self.dt = dt
        self.duration = duration

        self.loading = loading
        self.parameters = parameters
        self.L = self.parameters[-1]

        # Train passing 
        try:
            Kraftdefinition=pd.read_csv('TrainPassing/Inputdatei_1.txt',delim_whitespace=True)

            # # # % Koordinaten der Radsätze mit Bezug auf den ersten Radsatz [m]
            self.x_k=Kraftdefinition.iloc[:, 0].to_list()
            # # # % Achslasten in [N]
            self.P_k=Kraftdefinition.iloc[:, 1].to_list()
        except:
            pass

    def load_function(self, time_step):

        if (self.loading == "Pedestrian"):

            phi = 0
            if (time_step < self.L / self.parameters[1]):
                xp = self.parameters[1] * time_step
                phi = np.sin(np.pi * xp / self.L)
            else:
                phi = 0

            coeff_1 = 0.50
            coeff_2 = 0.10
            coeff_3 = 0.10

            Fn = (1+ coeff_1  * np.sin(2  *np.pi * self.parameters[0]* time_step)
                                    + coeff_2  * np.sin(4  *np.pi * self.parameters[0]* time_step -np.pi/2	)
                                     +coeff_3 * np.sin(6  *np.pi * self.parameters[0]* time_step -np.pi/2	)
            )  * self.parameters[2] * phi


            return Fn


        elif ( self.loading == "Harmonic"):
            return np.sin(np.sqrt(self.k/self.m)* time_step) * 1000


        elif (self.loading == "impact"):
            return self.m2 * 9.81
        
        elif (self.loading == "trainpassing"):
            self.v_train = self.parameters[1]
            time = np.arange(0, self.duration + self.dt, self.dt)
            nt = len(time)


            P_k_array = np.array(self.P_k)

            F_Mat = np.zeros((len(self.x_k)))        # rows -> Index of the train load, cols = number of time steps
                                                

            for j in range(len(self.x_k)):
                if ((-self.x_k[j] + self.v_train * time_step > 0) and (-self.x_k[j] + self.v_train * time_step < self.L)): 
                    xp =    -self.x_k[j] + self.v_train * time_step     # Condition, that the train needs to be on the bridge
                    phi = np.sin(np.pi * xp / self.L)
                    F_Mat[j] = self.P_k[j] * phi
                else:
                    F_Mat[j] = 0

            return F_Mat.sum(axis=0) 

        else:
            if time_step < 3e-3:
                return 1000
            else:
                return 0

    def solve(self):
        omega_0 = np.sqrt(self.k / self.m)
        omega_d = omega_0 * np.sqrt(1 - self.zeta**2)

        time = np.arange(0, self.duration + self.dt, self.dt)
        self.time_output = time
        u = np.zeros(len(time))

        ACum_i = 0
        BCum_i = 0

        for i, t in enumerate(time):
            if i > 0:
                y_i = np.exp(self.zeta * omega_0 * time[i]) * self.load_function(time[i]) * np.cos(omega_d * time[i])
                y_i_1 = np.exp(self.zeta * omega_0 * time[i-1]) * self.load_function(time[i-1]) * np.cos(omega_d * time[i-1])
                area_i = 0.5 * self.dt * (y_i + y_i_1)
                ACum_i += area_i

                y_i = np.exp(self.zeta * omega_0 * time[i]) * self.load_function(time[i]) * np.sin(omega_d * time[i])
                y_i_1 = np.exp(self.zeta * omega_0 * time[i-1]) * self.load_function(time[i-1]) * np.sin(omega_d * time[i-1])
                area_i = 0.5 * self.dt * (y_i + y_i_1)
                BCum_i += area_i

                u[i] = (1 / (self.m * omega_d)) * (ACum_i) * np.exp(-self.zeta * omega_0 * time[i]) * np.sin(omega_d * time[i]) - (1 / (self.m * omega_d)) * (BCum_i) * np.exp(-self.zeta * omega_0 * time[i]) * np.cos(omega_d * time[i])

        return time, u

    def calculate_velocity_and_acceleration(self):
        time, u = self.solve()
        velocity = (np.roll(u, -1) - np.roll(u, 1)) / (2 * self.dt)
        acceleration = (np.roll(u, -1) - 2 * u + np.roll(u, 1)) / (self.dt**2)
        # Remove the erroneous first and last elements
        velocity = velocity[1:-1]
        acceleration = acceleration[1:-1]
        time = time[1:-1]
        return time, velocity, acceleration

    def plot_solution(self):
        time, self.u = self.solve()
        self._time = time
        self.time_v_a , self.vel, self.acc = self.calculate_velocity_and_acceleration()


        fig,axes = plt.subplots(nrows = 3, ncols=1)

        axes[0].plot(self._time,self.u, label = "Deflection")
        axes[0].set_ylabel("Deflection in [m]")
        axes[1].plot(self.time_v_a, self.vel, label = "Velocity")
        axes[1].set_ylabel("Velocity in [m/s]")
        axes[2].plot(self.time_v_a, self.acc, label = "Acceleration")
        axes[2].set_ylabel("Acceleration in [m/s²]")
        fig.suptitle("Vibrations with the Duhamel integral")

        plt.xlabel('Time in [s]')
        plt.tight_layout()
        plt.grid(True)
        plt.show(block = False)
        plt.pause(10)
        plt.close()

    def plot_load_function(self):
        self.time_modal_load = np.arange(0, self.duration + self.dt, self.dt)
        self.load_values = np.array([self.load_function(t) for t in self.time_modal_load])

        plt.figure()
        plt.plot(self.time_modal_load ,self.load_values, label='Load Function')
        plt.xlabel('Time in [s]')
        plt.ylabel('Load in [N]')
        plt.title('Load Function over Time')
        plt.grid(True)
        plt.legend()
        plt.show()



# Impact = DuhamelSolver(1e6, 1000, 0.01, 0.001, 10, loading = "impact", parameters = [2,1.3,800,60], m2_imp = 100)
# Impact.plot_solution()


# u_stat = 100/1e6*9.81

# print("udyn", max(Impact.u))
# print("ustat", u_stat)
# print("faktor",  max(Impact.u)/u_stat )

v_train = 300/3.6
l_up = 25
f_train = v_train / l_up 
E = 2.1e11
I = 0.5
EI = E*I
l = 60
f = l**3/(48 * EI)
m = 0.5 * l * 5160

t_cal = 10


TrainPassing = DuhamelSolver(1/f, m , 0.0257,  0.001, t_cal, loading = "trainpassing", parameters = [f_train,v_train,0,l], m2_imp = 100)
TrainPassing.plot_solution()

# print("time for passing in [s]", (60+393.7) / v_train)


Schwingung = Balkenschwingungen(l, 1, 5063, 0.0257, E, I, t_cal,"Hinged-Hinged",None,l/2,300)



fig,axes = plt.subplots(nrows = 3, ncols=1)

axes[0].plot(TrainPassing._time,TrainPassing.u, label = "Deflection SDOF")
axes[0].plot(Schwingung.t,Schwingung.y, label = "Deflection continuum")
axes[0].set_ylabel("Deflection in [m]")
axes[1].plot(TrainPassing.time_v_a, TrainPassing.vel, label = "Velocity SDOF")
axes[1].plot(Schwingung.t,Schwingung.v, label = "Velocity continuum")
axes[1].set_ylabel("Velocity in [m/s]")
axes[2].plot(TrainPassing.time_v_a, TrainPassing.acc, label = "Acceleration SDOF")
axes[2].plot(Schwingung.t,Schwingung.a, label = "Acceleration continuum")
axes[2].set_ylabel("Acceleration in [m/s²]")
fig.suptitle("Vibrations with the Duhamel integral and continuum")

plt.xlabel('Time in [s]')
plt.tight_layout()
plt.grid(True)
plt.show(block = False)
plt.pause(10)
plt.close()



import numpy as np
class Baustoffe:
    def __init__(self, liste_baustoffe, h_c=0.13, b_c=4, traeger="FT", t0=1, t=365,ts = 1):
        """
        Klasse zur Berechnung der Baustoffparameter: Eingabe in [MN/m², m , MN, etc.]
        liste_baustoffe: Eingabe von [fck,fyk] \n
        h_c = Höhe des Betonträgers zur Berechnung von h0 in [m] \n
        b_c = Bezugsbreite in [m] \n
        traeger = Eingabe des Umfangs, welcher der freien Trocknung ausgesetzt ist \n
                   Für Fertigteile ergibt sich hier: h0 = 2*h   ("FT")
                   Für Ortbetonplatten folgt:        h0 = h     ("Ortbeton")
        t0 = Tatsächliches Betonalter beim Belastungsbeginn  \n
        t  = Betrachteter Zeitpunkt für die ermittelte Kriechzahl
        """
        # Teilsicherheitsbeiwerte für Baustoffe
        self.gamma_c = 1.5
        self.gamma_s = 1.15
        # Stahlbeton
        self.fck = liste_baustoffe[0]
        self.fcd = self.fck / self.gamma_c * 0.85
        self.fctm = 0.30 * self.fck ** (2 / 3)  # Nach Eurocode 2 Tabelle 3.1
        self.fctk_005 = 0.7 * self.fctm
        self.fcm = self.fck + 8
        self.Ecm = 22000 * (0.1 * self.fcm) ** (0.3)

        self.alpha_T = 1.2e-5

        # Kriechen und Schwinden

        self.traeger = traeger
        self.Zement = "I42.5N"
        self.LF = 80

        self.t0 = t0  # Annahme: Belastungsbeginn der Fertigteile nach 28 d
        self.t = t
        self.ts = ts

        # Querschnittswerte für Kriechen / Schwinden

        self.b_c = b_c
        self.h_c = h_c

        self.A = self.b_c * self.h_c
        if self.traeger == "FT":
            self.u = self.b_c
        elif self.traeger == "Normal":
            self.u = self.b_c * 2  # Zur Trocknung freigesetzter Umfang
        else:
            self.u = self.b_c * 2

        self.h_0 = 2 * self.A / self.u

        # Stahl
        self.fyk = liste_baustoffe[1]
        self.fyd = self.fyk / self.gamma_s
        self.Es = 2e5
        self.Materialgesetz_Beton()
        self.Kriechzahlen()
        self.Schwindzahlen()

    def Materialgesetz_Beton(self):
        self.eps_c1 = 0.7 * self.fcm**0.31
        self.eps_c1 = min(self.eps_c1, 2.8)  # Sicherstellen, dass eps_c1 <= 2.8

    def Kriechzahlen(self):
        self.Zement = "I42.5N"

        if self.Zement == "I42.5N" or self.Zement == "I32.5R":
            self.alpha = 0
        elif (
            self.Zement == "I42.5R"
            or self.Zement == "I52.5N"
            or self.Zement == "I52.5R"
        ):
            self.alpha = 1
        else:
            self.alpha = -1

        self.t_0 = self.t0

        self.t_0_eff = max(
            self.t_0 * (1 + 9 / (2 + self.t_0 ** (1.2))) ** self.alpha, 0.5
        )

        self.t_infty = (
            self.t
        )  # self.t_0+15       #70 * 365  Annahme: 15 Tage nach der Betonage

        self.RH = self.LF  # Außenliegendes Bauteil

        # Fallunterscheidung für Druckfestigkeit
        self.alpha_1 = min((35 / self.fcm) ** 0.7, 1)
        self.alpha_2 = min((35 / self.fcm) ** 0.2, 1)
        self.alpha_3 = min((35 / self.fcm) ** 0.5, 1)

        # Einfluss der Luftfeuchte und wirksamer Bauteildicke
        if self.fcm <= 35:
            self.phi_rh = 1 + (1 - self.RH / 100) / (
                0.1 * (self.h_0 * 1000) ** (0.3333333)
            )
        else:
            self.phi_rh = (
                1
                + (1 - self.RH / 100)
                / (0.1 * (self.h_0 * 1000) ** (0.3333333))
                * self.alpha_1
            ) * self.alpha_2

        # Einfluss der Betondruckfestigkeit
        self.beta_fcm = 16.8 / np.sqrt(self.fcm)
        # Einfluss des Belastungsbeginns
        self.beta_t0 = 1 / (0.1 + self.t_0_eff**0.2)

        # Einfluss der Luftfeuchte - Beta-Beiwert
        if self.fcm <= 35:
            self.beta_h = min(
                1.5 * (1 + (0.012 * self.RH) ** 18) * self.h_0 * 1000 + 250, 1500
            )
        else:
            self.beta_h = min(
                1.5 * (1 + (0.012 * self.RH) ** 18) * self.h_0 * 1000
                + 250 * self.alpha_3,
                1500 * self.alpha_3,
            )

        # Einfluss der Belastungsdauer und Belastungsbeginn

        self.beta_c_t_t0 = (
            (self.t_infty - self.t_0) / (self.beta_h + self.t_infty - self.t_0)
        ) ** 0.30

        self.phi_infty = (
            self.phi_rh * self.beta_fcm * self.beta_t0 * self.beta_c_t_t0
        )  # Kriechzahl zum Zeitpunkt t

        print("Kriechzahl phi ", self.phi_infty)

    def Schwindzahlen(self):
        self.beta_rh = 1.55 * (1 - (self.RH / 100) ** 3)

        if self.Zement == "I42.5N" or self.Zement == "I32.5R":
            self.alpha_as = 700
            self.alpha_ds1 = 4
            self.alpha_ds2 = 0.12
        elif (
            self.Zement == "I42.5R"
            or self.Zement == "I52.5N"
            or self.Zement == "I52.5R"
        ):
            self.alpha_as = 600
            self.alpha_ds1 = 6
            self.alpha_ds2 = 0.12
        else:
            self.alpha_as = 800
            self.alpha_ds1 = 3
            self.alpha_ds2 = 0.12

        self.epsilon_cd_0 = (
            0.85
            * ((220 + 110 * self.alpha_ds1) * np.exp(-self.alpha_ds2 * self.fcm / 10))
            * 1e-6
            * self.beta_rh
        )

        ts = self.ts
        
        t = self.t_infty
        self.t_infty_s = t

        self.beta_ds = (t - ts) / ((t - ts) + 0.04 * np.sqrt(self.h_0**3))

        if self.h_0 * 1000 <= 100:
            self.k_h = 1.00
        elif self.h_0 * 1000 > 100 and self.h_0 * 1000 <= 200:
            self.k_h = 1.00 - 0.15 / 100 * self.h_0 * 1000
        elif self.h_0 * 1000 > 200 and self.h_0 * 1000 <= 300:
            self.k_h = 0.85 - 0.10 / 100 * self.h_0 * 1000
        elif self.h_0 * 1000 > 300 and self.h_0 * 1000 <= 500:
            self.k_h = 0.75 - 0.05 / 100 * self.h_0 * 1000
        elif self.h_0 * 1000 > 500:
            self.k_h = 0.70

        self.epsilon_cd = self.beta_ds * self.epsilon_cd_0 * self.k_h

        # Autogenes Schwinden
        self.epsilon_ca_infty = 2.5 * (self.fck - 10) * 1e-6
        self.beta_as = 1 - np.exp(-0.2 * np.sqrt(t))

        self.epsilon_ca = self.beta_as * self.epsilon_ca_infty

        # Gesamtschwinden
        self.epsilon_cs = self.epsilon_cd + self.epsilon_ca

        print("Gesamtschwindmaß ", self.epsilon_cs)




from Meshing import *
_num_x = 10
_num_y = 10


A = MeshingClass(d1=10, d2=10, num_x=_num_x, num_y=_num_y, elem_type="rect"); A.generating_rectangular_mesh()

B = MeshingClass(d1=10, d2=10, num_x=_num_x, num_y=_num_y, elem_type="rect"); B.generating_rectangular_mesh()
B.NL[:,1] += A.NL[:,1].max()  # rechts an A

C = MeshingClass(d1=10, d2=10, num_x=_num_x, num_y=_num_y, elem_type="rect"); C.generating_rectangular_mesh()
C.NL[:,1] += A.NL[:,1].max()  # oben auf A
C.NL[:,0] += B.NL[:,0].max()

A.merge_with_mesh(B)
A.merge_with_mesh(C)

BottomNodes = A.get_bottom_border_nodes()
print(BottomNodes)

A.plot_mesh()







# ============================================================
# Plotting.py
# Reines Plotting-Modul für STABRAUM
# Erwartet ein AnalysisResults-Objekt:  res = calc.run()
# ============================================================

from __future__ import annotations

import numpy as np
import matplotlib.pyplot as plt

import matplotlib.patches as mpatches
from matplotlib.patches import FancyArrowPatch
from matplotlib.widgets import Slider
from mpl_toolkits.mplot3d.art3d import Line3DCollection


# ------------------------------------------------------------
# Hilfsfunktion: gleiche Achsenskalierung im 3D
# ------------------------------------------------------------
def set_axes_equal_3d(ax, extra: float = 0.0):
    x_limits = np.array(ax.get_xlim3d(), dtype=float)
    y_limits = np.array(ax.get_ylim3d(), dtype=float)
    z_limits = np.array(ax.get_zlim3d(), dtype=float)

    ranges = np.array([np.ptp(lim) for lim in (x_limits, y_limits, z_limits)], dtype=float)
    max_range = float(max(ranges.max(), 1e-9))

    mids = np.array([lim.mean() for lim in (x_limits, y_limits, z_limits)], dtype=float)
    half = (1.0 + float(extra)) * max_range / 2.0

    ax.set_xlim3d(mids[0] - half, mids[0] + half)
    ax.set_ylim3d(mids[1] - half, mids[1] + half)
    ax.set_zlim3d(mids[2] - half, mids[2] + half)

    try:
        ax.set_box_aspect((1, 1, 1))
    except Exception:
        pass


# ============================================================
# StructurePlotter
# ============================================================
class StructurePlotter:
    """
    Reines Plotting.
    Erwartet ein AnalysisResults-Objekt (res = calc.run()).
    """

    def __init__(self, res):
        self.res = res
        self.Inp = res.Inp

        self.nodes = self.Inp.nodes
        self.na = self.Inp.members["na"]
        self.ne = self.Inp.members["ne"]

    # --------------------------------------------------------
    # Geometrie-Helfer
    # --------------------------------------------------------
    def _pt(self, n: int) -> np.ndarray:
        return np.array(
            [
                float(self.nodes["x[m]"][n - 1]),
                float(self.nodes["y[m]"][n - 1]),
                float(self.nodes["z[m]"][n - 1]),
            ],
            dtype=float,
        )

    def _tangent(self, a: int, e: int):
        Pi, Pj = self._pt(int(a)), self._pt(int(e))
        v = Pj - Pi
        L = float(np.linalg.norm(v))
        if L < 1e-15:
            raise ValueError("Elementlänge ~ 0")
        return Pi, Pj, v / L, L

    def _stable_normal(self, t: np.ndarray, prefer="y") -> np.ndarray:
        axes = {
            "x": np.array([1.0, 0.0, 0.0]),
            "y": np.array([0.0, 1.0, 0.0]),
            "z": np.array([0.0, 0.0, 1.0]),
        }
        u = axes.get(prefer, axes["y"])
        if abs(float(np.dot(t, u))) > 0.95:
            u = axes["z"] if prefer != "z" else axes["x"]

        w = np.cross(t, u)
        n = float(np.linalg.norm(w))
        if n < 1e-15:
            raise ValueError("Kein Orthogonalvektor")
        return w / n

    def _orth_unit_2d(self, xi, zi, xj, zj) -> np.ndarray:
        """
        Orthogonaler Einheitsvektor zur Stabachse in x-z-Ebene.
        """
        v = np.array([float(xj - xi), 0.0, float(zj - zi)], dtype=float)
        y_unit = np.array([0.0, 1.0, 0.0], dtype=float)
        perp = np.cross(v, y_unit)[[0, 2]]
        n = float(np.linalg.norm(perp))
        if n < 1e-15:
            raise ValueError("Elementlänge ~ 0")
        return perp / n

    def _field_map(self):
        return {
            "N": self.res.N_el_i_store,
            "VY": self.res.VY_el_i_store,
            "VZ": self.res.VZ_el_i_store,
            "MX": self.res.MX_el_i_store,
            "MY": self.res.MY_el_i_store,
            "MZ": self.res.MZ_el_i_store,
        }

    # ========================================================
    # 2D Struktur (x-z) + Nummerierung
    # ========================================================
    def plot_structure_2d(self, node_labels: bool = False, elem_labels: bool = False):
        fig, ax = plt.subplots(figsize=(10, 6))

        # Stäbe zeichnen
        for idx, (a, e) in enumerate(zip(self.na, self.ne), start=1):
            Pi, Pj = self._pt(int(a)), self._pt(int(e))
            ax.plot([Pi[0], Pj[0]], [Pi[2], Pj[2]], color="black", lw=1.0)

            # Elementnummer am Mittelpunkt
            if elem_labels:
                xm = 0.5 * (Pi[0] + Pj[0])
                zm = 0.5 * (Pi[2] + Pj[2])
                ax.text(
                    xm,
                    zm,
                    f"E{idx}",
                    fontsize=9,
                    ha="center",
                    va="center",
                    bbox=dict(boxstyle="round,pad=0.15", fc="white", ec="gray", lw=0.5),
                )

        # Knotennummern
        if node_labels:
            n_nodes = len(self.nodes["x[m]"])
            for n in range(1, n_nodes + 1):
                P = self._pt(n)
                ax.plot(P[0], P[2], marker="o", markersize=3, color="black")
                ax.text(
                    P[0],
                    P[2],
                    f"N{n}",
                    fontsize=9,
                    ha="left",
                    va="bottom",
                    bbox=dict(boxstyle="round,pad=0.15", fc="white", ec="gray", lw=0.5),
                )

        ax.set_xlabel("X [m]")
        ax.set_ylabel("Z [m]")
        ax.set_title("Unverformte Struktur (x-z)")
        ax.grid(True)
        ax.set_aspect("equal", adjustable="datalim")
        ax.relim()
        ax.autoscale_view()
        return fig, ax

    # ========================================================
    # 2D Endwerte (statisch)
    # ========================================================
    def plot_endforces_2d(
        self,
        kind="MY",
        scale=5.0,
        invert_y=False,
        node_labels=False,
        elem_labels=False,
    ):
        fig, ax = self.plot_structure_2d(node_labels=node_labels, elem_labels=elem_labels)
        Q = self._field_map()[kind.upper()]  # (nElem,2,1)

        for i, (a, e) in enumerate(zip(self.na, self.ne)):
            Pi, Pj = self._pt(int(a)), self._pt(int(e))
            ix, iz = Pi[0], Pi[2]
            jx, jz = Pj[0], Pj[2]

            try:
                u = self._orth_unit_2d(ix, iz, jx, jz)
            except ValueError:
                continue

            qa = float(Q[i, 0, 0])
            qb = float(Q[i, 1, 0])

            def endpt(x, z, val):
                if abs(val) < 1e-15:
                    return x, z
                col = "blue" if val >= 0 else "red"
                link = 0.05 * float(scale)
                cx = x + link * u[0] * val * float(scale)
                cz = z + link * u[1] * val * float(scale)
                ax.plot([x, cx], [z, cz], color=col, lw=1)
                ax.text(cx, cz, f"{kind}={val:.3f}", color=col, fontsize=8)
                return cx, cz

            ca = endpt(ix, iz, qa)
            cb = endpt(jx, jz, qb)

            # Verbindungslinie immer zeichnen
            ax.plot([ca[0], cb[0]], [ca[1], cb[1]], color="black", lw=1)

        ax.legend(
            handles=[
                mpatches.Patch(color="blue", label=f"{kind} ≥ 0"),
                mpatches.Patch(color="red", label=f"{kind} < 0"),
            ]
        )

        if invert_y:
            ax.invert_yaxis()

        ax.relim()
        ax.autoscale_view()
        return fig, ax

    # ========================================================
    # 2D Endwerte (DYNAMISCH, Slider für scale)
    # ========================================================
    def plot_endforces_2d_interactive(
        self,
        kind="MY",
        scale_init=5.0,
        invert_y=False,
        node_labels=False,
        elem_labels=False,
        robust_ref=True,
    ):
        fig, ax = self.plot_structure_2d(node_labels=node_labels, elem_labels=elem_labels)
        Q = self._field_map()[kind.upper()]  # (nElem,2,1)

        # Referenz (nur für sinnvolle Slider-Grenzen)
        vals = np.abs(Q[:, :, 0]).ravel()
        if vals.size == 0:
            qref = 1.0
        else:
            qref = float(np.percentile(vals, 95)) if robust_ref else float(vals.max())
            qref = max(qref, 1e-12)

        smin = 0.0
        smax = max(float(scale_init) * 50.0, 10.0)

        ax.legend(
            handles=[
                mpatches.Patch(color="blue", label=f"{kind} ≥ 0"),
                mpatches.Patch(color="red", label=f"{kind} < 0"),
            ]
        )

        def _mark_dyn(artist):
            try:
                artist._dyn = True
            except Exception:
                pass
            return artist

        def _clear_dyn():
            for ln in list(getattr(ax, "lines", [])):
                if getattr(ln, "_dyn", False):
                    ln.remove()
            for p in list(getattr(ax, "patches", [])):
                if getattr(p, "_dyn", False):
                    p.remove()
            for t in list(getattr(ax, "texts", [])):
                if getattr(t, "_dyn", False):
                    t.remove()

        def draw(scale):
            _clear_dyn()
            scale = float(scale)

            for i, (a, e) in enumerate(zip(self.na, self.ne)):
                Pi, Pj = self._pt(int(a)), self._pt(int(e))
                ix, iz = Pi[0], Pi[2]
                jx, jz = Pj[0], Pj[2]

                try:
                    u = self._orth_unit_2d(ix, iz, jx, jz)
                except ValueError:
                    continue

                qa = float(Q[i, 0, 0])
                qb = float(Q[i, 1, 0])

                def endpt(x, z, val):
                    if abs(val) < 1e-15:
                        return x, z
                    col = "blue" if val >= 0 else "red"
                    link = 0.05 * scale
                    cx = x + link * u[0] * val * scale
                    cz = z + link * u[1] * val * scale
                    ln = ax.plot([x, cx], [z, cz], color=col, lw=1)[0]
                    _mark_dyn(ln)
                    txt = ax.text(cx, cz, f"{kind}={val:.3f}", color=col, fontsize=8)
                    _mark_dyn(txt)
                    return cx, cz

                ca = endpt(ix, iz, qa)
                cb = endpt(jx, jz, qb)

                ln2 = ax.plot([ca[0], cb[0]], [ca[1], cb[1]], color="black", lw=1)[0]
                _mark_dyn(ln2)

            ax.relim()
            ax.autoscale_view()
            ax.set_aspect("equal", adjustable="datalim")
            if invert_y:
                ax.invert_yaxis()
            fig.canvas.draw_idle()

        fig.subplots_adjust(bottom=0.20)
        ax_scale = fig.add_axes([0.15, 0.08, 0.70, 0.03])
        s_scale = Slider(ax_scale, "Scale", smin, smax, valinit=float(scale_init))

        s_scale.on_changed(lambda _: draw(s_scale.val))
        draw(scale_init)

        return fig, ax, s_scale

    # ========================================================
    # 3D verformte Struktur (interaktiv) + Nummerierung
    # ========================================================
    def plot_structure_deformed_3d_interactive(
        self,
        scale_init=1.0,
        show_undeformed=True,
        node_labels=False,
        elem_labels=False,
    ):
        def ux(n): return float(self.res.u_ges[7 * (n - 1) + 0])
        def uy(n): return float(self.res.u_ges[7 * (n - 1) + 1])
        def uz(n): return float(self.res.u_ges[7 * (n - 1) + 3])

        fig = plt.figure(figsize=(9, 7))
        ax = fig.add_subplot(projection="3d")

        # unverformte Stäbe
        if show_undeformed:
            segs = []
            for a, e in zip(self.na, self.ne):
                segs.append([self._pt(int(a)), self._pt(int(e))])
            ax.add_collection3d(Line3DCollection(segs, colors="lightgray", linewidths=1, zorder=0))

        # deformierte Linien
        deformed_lines = []
        for a, e in zip(self.na, self.ne):
            a, e = int(a), int(e)
            Pi, Pj = self._pt(a), self._pt(e)

            xd = [Pi[0] + float(scale_init) * ux(a), Pj[0] + float(scale_init) * ux(e)]
            yd = [Pi[1] + float(scale_init) * uy(a), Pj[1] + float(scale_init) * uy(e)]
            zd = [Pi[2] + float(scale_init) * uz(a), Pj[2] + float(scale_init) * uz(e)]
            (ld,) = ax.plot(xd, yd, zd, lw=2, zorder=3)
            deformed_lines.append((a, e, ld))

        # Texte
        node_texts = []
        elem_texts = []

        if node_labels:
            n_nodes = len(self.nodes["x[m]"])
            for n in range(1, n_nodes + 1):
                P = self._pt(n)
                txt = ax.text(
                    P[0] + float(scale_init) * ux(n),
                    P[1] + float(scale_init) * uy(n),
                    P[2] + float(scale_init) * uz(n),
                    f"N{n}",
                    fontsize=8,
                    zorder=4,
                )
                node_texts.append((n, txt))

        if elem_labels:
            for idx, (a, e) in enumerate(zip(self.na, self.ne), start=1):
                a, e = int(a), int(e)
                Pi, Pj = self._pt(a), self._pt(e)
                xm = 0.5 * (Pi[0] + Pj[0]) + float(scale_init) * 0.5 * (ux(a) + ux(e))
                ym = 0.5 * (Pi[1] + Pj[1]) + float(scale_init) * 0.5 * (uy(a) + uy(e))
                zm = 0.5 * (Pi[2] + Pj[2]) + float(scale_init) * 0.5 * (uz(a) + uz(e))
                txt = ax.text(xm, ym, zm, f"E{idx}", fontsize=8, zorder=4)
                elem_texts.append((idx, a, e, txt))

        ax.set_xlabel("X [m]")
        ax.set_ylabel("Y [m]")
        ax.set_zlabel("Z [m]")
        ax.set_title("Verformte Struktur 3D (interaktiv)")

        set_axes_equal_3d(ax, extra=0.05)

        fig.subplots_adjust(bottom=0.18)
        ax_scale = fig.add_axes([0.15, 0.08, 0.7, 0.03])
        s_scale = Slider(ax_scale, "Scale", 0.0, float(scale_init) * 10000.0, valinit=float(scale_init))

        def update(_):
            s = float(s_scale.val)

            for a, e, line in deformed_lines:
                Pi, Pj = self._pt(a), self._pt(e)
                line.set_data_3d(
                    [Pi[0] + s * ux(a), Pj[0] + s * ux(e)],
                    [Pi[1] + s * uy(a), Pj[1] + s * uy(e)],
                    [Pi[2] + s * uz(a), Pj[2] + s * uz(e)],
                )

            for n, txt in node_texts:
                P = self._pt(n)
                txt.set_position((P[0] + s * ux(n), P[1] + s * uy(n)))
                txt.set_3d_properties(P[2] + s * uz(n), zdir="z")

            for idx, a, e, txt in elem_texts:
                Pi, Pj = self._pt(a), self._pt(e)
                xm = 0.5 * (Pi[0] + Pj[0]) + s * 0.5 * (ux(a) + ux(e))
                ym = 0.5 * (Pi[1] + Pj[1]) + s * 0.5 * (uy(a) + uy(e))
                zm = 0.5 * (Pi[2] + Pj[2]) + s * 0.5 * (uz(a) + uz(e))
                txt.set_position((xm, ym))
                txt.set_3d_properties(zm, zdir="z")

            set_axes_equal_3d(ax, extra=0.05)
            fig.canvas.draw_idle()

        s_scale.on_changed(update)
        update(None)
        return fig, ax, s_scale

    # ========================================================
    # 3D Schnittgrößen-Diagramm entlang der Stäbe (DYNAMISCH)
    # ========================================================
    def plot_diagram_3d_interactive(
        self,
        kind="MY",
        n_stations=30,
        scale_init=1.0,
        prefer_axis="y",
        show_structure=True,
        robust_ref=True,
        title=None,
    ):
        fig = plt.figure(figsize=(9, 7))
        ax = fig.add_subplot(projection="3d")
        Q = self._field_map()[kind.upper()]  # (nElem,2,1)

        vals = np.abs(Q[:, :, 0]).ravel()
        qref = float(np.percentile(vals, 95)) if (robust_ref and vals.size) else float(vals.max() if vals.size else 1.0)
        qref = max(qref, 1e-12)

        if show_structure:
            segs = []
            for a, e in zip(self.na, self.ne):
                segs.append([self._pt(int(a)), self._pt(int(e))])
            ax.add_collection3d(Line3DCollection(segs, colors="lightgray", linewidths=1, zorder=0))

        diagram_lines = []
        elem_cache = []
        prev_w = None

        for i, (a, e) in enumerate(zip(self.na, self.ne)):
            a, e = int(a), int(e)
            Pi, Pj, t, L = self._tangent(a, e)
            w = self._stable_normal(t, prefer=prefer_axis)
            if prev_w is not None and float(np.dot(w, prev_w)) < 0.0:
                w = -w
            prev_w = w

            qa = float(Q[i, 0, 0])
            qb = float(Q[i, 1, 0])

            s = np.linspace(0.0, L, int(n_stations))
            q = np.linspace(qa, qb, int(n_stations))
            elem_cache.append({"Pi": Pi, "t": t, "w": w, "s": s, "q": q})

            alpha0 = float(scale_init) / qref
            pts0 = Pi[None, :] + s[:, None] * t[None, :] + (alpha0 * q)[:, None] * w[None, :]
            (ln,) = ax.plot(pts0[:, 0], pts0[:, 1], pts0[:, 2], lw=2, zorder=3)
            diagram_lines.append(ln)

        ax.set_xlabel("X [m]")
        ax.set_ylabel("Y [m]")
        ax.set_zlabel("Z [m]")
        ax.set_title(title or f"{kind}-Diagramm 3D (interaktiv)  |  ref={qref:.3g}")
        set_axes_equal_3d(ax, extra=0.05)

        fig.subplots_adjust(bottom=0.18)
        ax_scale = fig.add_axes([0.15, 0.08, 0.7, 0.03])
        s_scale = Slider(ax_scale, "Scale", 0.0, float(scale_init) * 200.0, valinit=float(scale_init))

        def update(_):
            scale = float(s_scale.val)
            alpha = scale / qref
            for ln, cache in zip(diagram_lines, elem_cache):
                Pi = cache["Pi"]
                t = cache["t"]
                w = cache["w"]
                s = cache["s"]
                q = cache["q"]
                pts = Pi[None, :] + s[:, None] * t[None, :] + (alpha * q)[:, None] * w[None, :]
                ln.set_data_3d(pts[:, 0], pts[:, 1], pts[:, 2])

            set_axes_equal_3d(ax, extra=0.05)
            fig.canvas.draw_idle()

        s_scale.on_changed(update)
        update(None)
        return fig, ax, s_scale

    # ========================================================
    # Robust: dynamische Artists markieren/aufräumen (global)
    # ========================================================
    def _mark_dyn(self, artist):
        try:
            artist._dyn = True
        except Exception:
            pass
        return artist

    def _clear_dyn(self, ax):
        for coll in list(getattr(ax, "collections", [])):
            if getattr(coll, "_dyn", False):
                coll.remove()
        for ln in list(getattr(ax, "lines", [])):
            if getattr(ln, "_dyn", False):
                ln.remove()
        for p in list(getattr(ax, "patches", [])):
            if getattr(p, "_dyn", False):
                p.remove()
        for t in list(getattr(ax, "texts", [])):
            if getattr(t, "_dyn", False):
                t.remove()

    # ========================================================
    # Auflagerreaktionen: r = K u - F
    # ========================================================
    def _reaction_vector(self) -> np.ndarray:
        return np.asarray(self.res.GesMat @ self.res.u_ges - self.res.FGes, dtype=float)

    def _draw_force_arrow(self, ax, x, z, dx, dz, color="green", lw=2):
        arr = FancyArrowPatch(
            (x, z),
            (x + float(dx), z + float(dz)),
            arrowstyle="-|>",
            mutation_scale=12,
            lw=lw,
            color=color,
        )
        ax.add_patch(arr)
        return self._mark_dyn(arr)

    def _draw_moment_double_arrow(self, ax, x, z, m, radius, color="purple", lw=2):
        sign = 1.0 if float(m) >= 0.0 else -1.0
        r = float(radius)

        a1 = FancyArrowPatch(
            (x - r, z),
            (x + r, z),
            connectionstyle=f"arc3,rad={0.6*sign}",
            arrowstyle="<->",
            mutation_scale=12,
            lw=lw,
            color=color,
        )
        ax.add_patch(a1)
        return self._mark_dyn(a1)

    def _support_nodes(self):
        dfR = getattr(self.Inp, "RestraintData", None)
        if dfR is None:
            return []
        if "Node" not in dfR.columns:
            return []
        return sorted(set(int(n) for n in dfR["Node"].values))

    def _length_ref_xz(self, frac=0.03):
        xs = np.asarray(self.nodes["x[m]"], dtype=float)
        zs = np.asarray(self.nodes["z[m]"], dtype=float)
        span = max(float(xs.max() - xs.min()), float(zs.max() - zs.min()), 1e-9)
        return float(frac) * span

    # ========================================================
    # Auflagerreaktionen 2D (interaktiv) + Tabelle + Lasten
    # ========================================================
    def plot_support_reactions_table_2d(
        self,
        moment_components=("Mx", "My", "Mz"),
        title="Auflagerreaktionen (Tabelle)",
        fmt_force="{:+.3f}",
        fmt_moment="{:+.3f}",
    ):
        r = self._reaction_vector()
        support_nodes = self._support_nodes()

        if not support_nodes:
            fig, ax = plt.subplots(figsize=(8, 3))
            ax.axis("off")
            ax.text(0.5, 0.5, "Keine Auflagerknoten in RestraintData gefunden.", ha="center", va="center")
            return fig, ax

        mom_set = set(m.upper() for m in moment_components)
        show_Mx = "MX" in mom_set
        show_My = "MY" in mom_set
        show_Mz = "MZ" in mom_set

        cols = ["Node", "Rx [MN]", "Rz [MN]"]
        if show_Mx: cols.append("Mx [MNm]")
        if show_My: cols.append("My [MNm]")
        if show_Mz: cols.append("Mz [MNm]")

        rows = []
        for n in support_nodes:
            gdof = 7 * (n - 1)
            Rx = float(r[gdof + 0])
            Rz = float(r[gdof + 3])
            Mz = float(r[gdof + 2])
            My = float(r[gdof + 4])
            Mx = float(r[gdof + 5])

            row = [str(n), fmt_force.format(Rx), fmt_force.format(Rz)]
            if show_Mx: row.append(fmt_moment.format(Mx))
            if show_My: row.append(fmt_moment.format(My))
            if show_Mz: row.append(fmt_moment.format(Mz))
            rows.append(row)

        nrows = len(rows) + 1
        fig_h = max(2.2, 0.35 * nrows)
        fig, ax = plt.subplots(figsize=(10, fig_h))
        ax.axis("off")
        ax.set_title(title)

        table = ax.table(cellText=rows, colLabels=cols, loc="center", cellLoc="center")
        table.auto_set_font_size(False)
        table.set_fontsize(9)
        table.scale(1.0, 1.2)

        return fig, ax

    def plot_support_reactions_2d_interactive(
        self,
        invert_y=False,
        node_labels=True,
        elem_labels=False,
        show_forces=True,
        show_moments=True,
        scale_force_init=0.8,
        moment_radius_init=0.08,
        moment_scale_init=1.0,
        moment_kind_prefer="MY",
        robust_ref=True,
        Lref_frac=0.03,
        slider_force_max=10.0,
    ):
        fig, ax = self.plot_structure_2d(node_labels=node_labels, elem_labels=elem_labels)
        ax.set_title("Auflagerreaktionen (interaktiv)")

        r = self._reaction_vector()
        support_nodes = self._support_nodes()
        if not support_nodes:
            ax.text(0.5, 0.5, "Keine Auflagerknoten in RestraintData gefunden.", transform=ax.transAxes,
                    ha="center", va="center")
            return fig, ax, None

        Lref = self._length_ref_xz(frac=Lref_frac)

        Fvals = []
        for n in support_nodes:
            gdof = 7 * (n - 1)
            Fvals += [abs(float(r[gdof + 0])), abs(float(r[gdof + 3]))]
        Fvals = np.asarray(Fvals, dtype=float)
        if Fvals.size == 0:
            Fref = 1.0
        else:
            Fref = float(np.percentile(Fvals, 95)) if robust_ref else float(Fvals.max())
            Fref = max(Fref, 1e-12)

        def pick_moment_components(gdof_base):
            Mz = float(r[gdof_base + 2])
            My = float(r[gdof_base + 4])
            Mx = float(r[gdof_base + 5])
            return Mx, My, Mz

        def choose_moment(Mx, My, Mz):
            pref = moment_kind_prefer.upper()
            if pref == "MX": return Mx, "Mx"
            if pref == "MZ": return Mz, "Mz"
            return My, "My"

        def draw(scale_force, moment_radius, moment_scale):
            self._clear_dyn(ax)

            scale_force = float(scale_force)
            moment_radius = float(moment_radius)
            moment_scale = float(moment_scale)

            alphaF = (scale_force * Lref) / Fref

            for n in support_nodes:
                P = self._pt(int(n))
                x, z = float(P[0]), float(P[2])
                gdof = 7 * (int(n) - 1)

                Rx = float(r[gdof + 0])
                Rz = float(r[gdof + 3])

                Mx, My, Mz = pick_moment_components(gdof)
                Mplot, Mlab = choose_moment(Mx, My, Mz)

                if show_forces and (abs(Rx) > 1e-15 or abs(Rz) > 1e-15):
                    self._draw_force_arrow(ax, x, z, alphaF * Rx, alphaF * Rz, color="green")
                    self._mark_dyn(ax.text(x, z, f"R{n}", fontsize=8, color="green", ha="right", va="top"))
                    self._mark_dyn(ax.text(x, z, f"\nRx={Rx:+.3f} MN\nRz={Rz:+.3f} MN",
                                           fontsize=8, color="green", ha="left", va="top"))

                if show_moments and abs(Mplot) > 1e-15:
                    rr = moment_radius * moment_scale
                    self._draw_moment_double_arrow(ax, x, z, Mplot, radius=rr, color="purple")
                    self._mark_dyn(ax.text(x + rr, z + rr, f"{Mlab}={Mplot:+.3f} MNm",
                                           fontsize=8, color="purple"))

            ax.relim()
            ax.autoscale_view()
            ax.set_aspect("equal", adjustable="datalim")
            if invert_y:
                ax.invert_yaxis()
            fig.canvas.draw_idle()

        fig.subplots_adjust(bottom=0.25)

        ax_sF = fig.add_axes([0.15, 0.14, 0.70, 0.03])
        s_force = Slider(ax_sF, "Scale Force", 0.0, float(slider_force_max), valinit=float(scale_force_init))

        ax_sR = fig.add_axes([0.15, 0.09, 0.70, 0.03])
        s_rad = Slider(ax_sR, "Moment Radius", 0.0, float(moment_radius_init) * 20.0, valinit=float(moment_radius_init))

        ax_sM = fig.add_axes([0.15, 0.04, 0.70, 0.03])
        s_msc = Slider(ax_sM, "Moment Scale", 0.0, float(moment_scale_init) * 20.0, valinit=float(moment_scale_init))

        def update(_):
            draw(s_force.val, s_rad.val, s_msc.val)

        s_force.on_changed(update)
        s_rad.on_changed(update)
        s_msc.on_changed(update)

        draw(scale_force_init, moment_radius_init, moment_scale_init)
        return fig, ax, (s_force, s_rad, s_msc)

    def plot_nodal_loads_2d_interactive(
        self,
        invert_y=False,
        node_labels=True,
        elem_labels=False,
        show_forces=True,
        show_moments=True,
        scale_force_init=0.8,
        moment_radius_init=0.08,
        moment_scale_init=1.0,
        moment_kind_prefer="MY",
        robust_ref=True,
        Lref_frac=0.03,
        slider_force_max=10.0,
    ):
        fig, ax = self.plot_structure_2d(node_labels=node_labels, elem_labels=elem_labels)
        ax.set_title("Knotenlasten (interaktiv) – aus NodalForces.csv")

        # --- Lasttabelle holen (EXAKT dein Input) ---
        dfL = self.Inp.NodalForces 

        print(dfL)

        if dfL is None:
            # falls du sie doch "NodalForces" genannt hast:
            dfL = getattr(self.Inp, "NodalForces", None)

        if dfL is None:
            ax.text(0.5, 0.5, "Keine NodalForces Tabelle in Inp gefunden.", transform=ax.transAxes,
                    ha="center", va="center")
            return fig, ax, None

        # --- in Node -> FX,FY,FZ,MX,MY,MZ aggregieren (langes Format) ---
        data = {}  # node -> dict

        for _, row in dfL.iterrows():
            n   = int(row["Node"])
            dof = str(row["Dof"]).strip().upper()
            val = float(row["Value[MN/MNm]"])

            if n not in data:
                data[n] = {"FX": 0.0, "FY": 0.0, "FZ": 0.0, "MX": 0.0, "MY": 0.0, "MZ": 0.0}

            if dof in data[n]:
                data[n][dof] += val
            else:
                print(f"Warnung: unbekannter Dof '{dof}' in NodalForces (Node {n})")


        nodes = sorted(data.keys())
        Fx = np.array([data[n]["FX"] for n in nodes], dtype=float)
        Fz = np.array([data[n]["FZ"] for n in nodes], dtype=float)
        Mx = np.array([data[n]["MX"] for n in nodes], dtype=float)
        My = np.array([data[n]["MY"] for n in nodes], dtype=float)
        Mz = np.array([data[n]["MZ"] for n in nodes], dtype=float)

        print("nodes:", nodes)
        print("Fx:", Fx)
        print("Fz:", Fz)
        # --- Skalierung ---
        Lref = self._length_ref_xz(frac=Lref_frac)
        Fabs = np.hstack([np.abs(Fx), np.abs(Fz)])
        if Fabs.size == 0:
            Fref = 1.0
        else:
            Fref = float(np.percentile(Fabs, 95)) if robust_ref else float(Fabs.max())
            Fref = max(Fref, 1e-12)

        def choose_moment(mx, my, mz):
            pref = str(moment_kind_prefer).upper()
            if pref == "MX":
                return float(mx), "Mx"
            if pref == "MZ":
                return float(mz), "Mz"
            return float(my), "My"

        def draw(scale_force, moment_radius, moment_scale):
            self._clear_dyn(ax)

            scale_force = float(scale_force)
            moment_radius = float(moment_radius)
            moment_scale = float(moment_scale)

            alphaF = (scale_force * Lref) / Fref

            for n, fx, fz, mx, my, mz in zip(nodes, Fx, Fz, Mx, My, Mz):
                P = self._pt(int(n))
                x, z = float(P[0]), float(P[2])

                # Kräfte
                if show_forces and (abs(fx) > 1e-15 or abs(fz) > 1e-15):
                    self._draw_force_arrow(ax, x, z, alphaF * fx, alphaF * fz, color="orange")
                    self._mark_dyn(ax.text(x, z, f"F{n}", fontsize=8, color="orange", ha="right", va="top"))
                    self._mark_dyn(ax.text(x, z, f"\nFx={fx:+.3f} MN\nFz={fz:+.3f} MN",
                                        fontsize=8, color="orange", ha="left", va="top"))

                # Momente
                Mplot, Mlab = choose_moment(mx, my, mz)
                if show_moments and abs(Mplot) > 1e-15:
                    rr = moment_radius * moment_scale
                    self._draw_moment_double_arrow(ax, x, z, Mplot, radius=rr, color="brown")
                    self._mark_dyn(ax.text(x + rr, z + rr, f"{Mlab}={Mplot:+.3f} MNm",
                                        fontsize=8, color="brown"))

            ax.relim()
            ax.autoscale_view()
            ax.set_aspect("equal", adjustable="datalim")
            if invert_y:
                ax.invert_yaxis()
            fig.canvas.draw_idle()

        # --- Slider ---
        fig.subplots_adjust(bottom=0.25)

        ax_sF = fig.add_axes([0.15, 0.14, 0.70, 0.03])
        s_force = Slider(ax_sF, "Scale Force", 0.0, float(slider_force_max), valinit=float(scale_force_init))

        ax_sR = fig.add_axes([0.15, 0.09, 0.70, 0.03])
        s_rad = Slider(ax_sR, "Moment Radius", 0.0, float(moment_radius_init) * 20.0, valinit=float(moment_radius_init))

        ax_sM = fig.add_axes([0.15, 0.04, 0.70, 0.03])
        s_msc = Slider(ax_sM, "Moment Scale", 0.0, float(moment_scale_init) * 20.0, valinit=float(moment_scale_init))

        def update(_):
            draw(s_force.val, s_rad.val, s_msc.val)

        s_force.on_changed(update)
        s_rad.on_changed(update)
        s_msc.on_changed(update)

        draw(scale_force_init, moment_radius_init, moment_scale_init)
        return fig, ax, (s_force, s_rad, s_msc)




# STABRAUM-Dependencies
try:
    from Steifigkeitsmatrix import *
    from InputData import Input
    from results import AnalysisResults
except:
    from KIB_LAP.STABRAUM.Steifigkeitsmatrix import *
    from KIB_LAP.STABRAUM.InputData import Input
    from KIB_LAP.STABRAUM.results import AnalysisResults


import sympy as sp

import matplotlib.pyplot as plt
import numpy as np

from mpl_toolkits.mplot3d import Axes3D  # nötig für 3D-Plots
import matplotlib.patches as mpatches
from matplotlib.widgets import Slider
from mpl_toolkits.mplot3d.art3d import Line3DCollection


def _set_axes_equal(ax, extra: float = 0.0):
    """
    Erzwingt identische numerische Achsenlimits.
    optional: 'extra' = zusätzlicher Rand als Prozentsatz (0-1).
    """
    import numpy as np

    # aktuelle Grenzen holen
    x_limits = np.array(ax.get_xlim3d())
    y_limits = np.array(ax.get_ylim3d())
    z_limits = np.array(ax.get_zlim3d())

    # Spannweiten & gemeinsames Maximum
    ranges = np.array([np.ptp(lim) for lim in (x_limits, y_limits, z_limits)])
    max_range = ranges.max()

    # sind alle Punkte (fast) in einer Ebene? -> Mindestspanne ansetzen
    if max_range == 0:
        max_range = 1.0  # beliebiger Würfel von 1 m

    # Mittelpunkt­koordinaten
    mids = np.array([lim.mean() for lim in (x_limits, y_limits, z_limits)])

    half = (1 + extra) * max_range / 2
    ax.set_xlim3d(mids[0] - half, mids[0] + half)
    ax.set_ylim3d(mids[1] - half, mids[1] + half)
    ax.set_zlim3d(mids[2] - half, mids[2] + half)

    # Darstellungswürfel in aktuellen MPL-Versionen
    try:
        ax.set_box_aspect((1, 1, 1))
    except AttributeError:
        pass


class mainloop:
    def __init__(self):
        ##________ Subclasses __________##
        self.Inp = Input()
        self.ElemStem = ElemStema()
        ##_____ Class variables ___________##
        self.K_el_i_store = np.zeros((len(self.Inp.members), 14, 14))

        self.MY_el_i_store = np.zeros((len(self.Inp.members), 2, 1))
        self.VZ_el_i_store = np.zeros((len(self.Inp.members), 2, 1))

        self.MZ_el_i_store = np.zeros((len(self.Inp.members), 2, 1))
        self.VY_el_i_store = np.zeros((len(self.Inp.members), 2, 1))
        self.MX_el_i_store = np.zeros((len(self.Inp.members), 2, 1))

        self.MTP_el_i_store = np.zeros((len(self.Inp.members), 2, 1))
        self.MTS_el_i_store = np.zeros((len(self.Inp.members), 2, 1))
        self.N_el_i_store = np.zeros((len(self.Inp.members), 2, 1))

    # ------------------------------------------------------------
    # DEBUG / SINGULARITÄTSTEST
    # ------------------------------------------------------------
    def check_singularity(self, K=None, tol_row=1e-10, tol_rank=1e-8, verbose=True):
        """
        Prüft typische Ursachen für Singularität:
        - Rank-Defizit
        - Nullzeilen/Nullspalten (DOFs ohne Steifigkeit)
        - isolierte Knoten (nicht in members)
        - 0-Längen-Elemente
        Gibt ein Dict mit Diagnosen zurück.
        """
        if K is None:
            K = self.GesMat

        info = {}

        # -------- Rank --------
        try:
            r = np.linalg.matrix_rank(K, tol=tol_rank)
        except TypeError:
            # falls numpy Version keine tol als kw hat
            r = np.linalg.matrix_rank(K)
        n = K.shape[0]
        info["rank"] = int(r)
        info["n"] = int(n)
        info["deficit"] = int(n - r)

        # -------- Nullzeilen/Nullspalten --------
        row_norm = np.linalg.norm(K, axis=1)
        col_norm = np.linalg.norm(K, axis=0)
        zero_rows = np.where(row_norm < tol_row)[0]
        zero_cols = np.where(col_norm < tol_row)[0]

        info["zero_rows"] = zero_rows.tolist()
        info["zero_cols"] = zero_cols.tolist()

        # -------- DOF Mapping (nur wenn 7 dof/node) --------
        def gdof_to_node_ldof(gdof):
            return int(gdof // 7 + 1), int(gdof % 7)

        info["zero_rows_nodes"] = [gdof_to_node_ldof(i) for i in zero_rows]
        info["zero_cols_nodes"] = [gdof_to_node_ldof(i) for i in zero_cols]

        # -------- isolierte Knoten --------
        try:
            na = list(self.Inp.members["na"])
            ne = list(self.Inp.members["ne"])
            used = set(na) | set(ne)
            all_nodes = set(range(1, len(self.Inp.nodes["x[m]"]) + 1))
            isolated = sorted(all_nodes - used)
        except Exception:
            isolated = []
        info["isolated_nodes"] = isolated

        # -------- 0-Längen-Elemente --------
        zero_len_elems = []
        try:
            for eidx, (a, e) in enumerate(zip(self.Inp.members["na"], self.Inp.members["ne"]), start=1):
                xa = self.Inp.nodes["x[m]"][a-1]; ya = self.Inp.nodes["y[m]"][a-1]; za = self.Inp.nodes["z[m]"][a-1]
                xb = self.Inp.nodes["x[m]"][e-1]; yb = self.Inp.nodes["y[m]"][e-1]; zb = self.Inp.nodes["z[m]"][e-1]
                L = float(np.sqrt((xa-xb)**2 + (ya-yb)**2 + (za-zb)**2))
                if L < 1e-12:
                    zero_len_elems.append((eidx, int(a), int(e)))
        except Exception:
            pass
        info["zero_length_elements"] = zero_len_elems

        # -------- Ausgabe --------
        if verbose:
            print("\n=== Singularity check ===")
            print(f"size n = {n}, rank = {r}, deficit = {n-r}")
            print(f"zero rows: {len(zero_rows)}, zero cols: {len(zero_cols)}")
            if len(zero_rows) > 0:
                print("first zero rows (gdof -> node,ldof):")
                for gdof in zero_rows[:30]:
                    node, ldof = gdof_to_node_ldof(gdof)
                    print(f"  gdof {gdof:4d} -> node {node}, ldof {ldof}")
            if len(isolated) > 0:
                print("isolated nodes:", isolated)
            if len(zero_len_elems) > 0:
                print("zero-length elements (eid, na, ne):", zero_len_elems)

        return info



    def CalculateTransMat(self):
        print("Calculate Transmatrices")

        TransformationMatrices = np.zeros((len(self.Inp.members), 14, 14))
        na_memb = self.Inp.members["na"]
        ne_memb = self.Inp.members["ne"]
        for i in range(len(self.Inp.members["na"])):

            node_i = na_memb[i]  # Node number for node i of this member
            node_j = ne_memb[i]  # Node number for node j of this member

            # Index of DoF for this member
            ia = 2 * node_i - 2  # horizontal DoF at node i of this member
            ib = 2 * node_i - 1  # vertical DoF at node i of this member
            ja = 2 * node_j - 2  # horizontal DoF at node j of this member
            jb = 2 * node_j - 1  # vertical DoF at node j of this member

            # New positions = initial pos + cum deflection
            ix = self.Inp.nodes["x[m]"][node_i - 1]
            iy = self.Inp.nodes["y[m]"][node_i - 1]
            iz = self.Inp.nodes["z[m]"][node_i - 1]

            jx = self.Inp.nodes["x[m]"][node_j - 1]
            jy = self.Inp.nodes["y[m]"][node_j - 1]
            jz = self.Inp.nodes["z[m]"][node_j - 1]

            TM = self.ElemStem.TransformationMatrix([ix, iy, iz], [jx, jy, jz])

            TransformationMatrices[i, :, :] = TM
        print("Transmat")
        print(TransformationMatrices[0])
        return TransformationMatrices

    def BuildStructureStiffnessMatrix(self):
        """
        Standard construction of Primary and Structure stiffness matrix
        Construction of non-linear element stiffness matrix handled in a child function
        """
        Kp = np.zeros(
            [self.Inp.nDoF, self.Inp.nDoF]
        )  # Initialise the primary stiffness matrix
        self.member_length = []

        na_memb = self.Inp.members["na"]
        ne_memb = self.Inp.members["ne"]
        crosssec_members = self.Inp.members["cs"]

        for i in range(0, len(self.Inp.members["na"]), 1):
            node_i = na_memb[i]  # Node number for node i of this member
            node_j = ne_memb[i]  # Node number for node j of this member

            # New positions = initial pos + cum deflection
            ix = self.Inp.nodes["x[m]"][node_i - 1]
            iy = self.Inp.nodes["y[m]"][node_i - 1]
            iz = self.Inp.nodes["z[m]"][node_i - 1]

            jx = self.Inp.nodes["x[m]"][node_j - 1]
            jy = self.Inp.nodes["y[m]"][node_j - 1]
            jz = self.Inp.nodes["z[m]"][node_j - 1]

            dx = abs(ix - jx)
            dy = abs(iy - jy)
            dz = abs(iz - jz)

            length = np.sqrt(dx**2 + dy**2 + dz**2)
            self.member_length.append(length)

            num_cs = crosssec_members[i]
            mat_num_i = self.Inp.CrossSection.loc[
                self.Inp.CrossSection["No"] == num_cs, "material"
            ].iloc[0]

            I_y_i = self.Inp.CrossSection.loc[
                self.Inp.CrossSection["No"] == num_cs, "Iy"
            ].iloc[0]
            I_z_i = self.Inp.CrossSection.loc[
                self.Inp.CrossSection["No"] == num_cs, "Iz"
            ].iloc[0]
            A_i = self.Inp.CrossSection.loc[
                self.Inp.CrossSection["No"] == num_cs, "A"
            ].iloc[0]
            I_w_i = self.Inp.CrossSection.loc[
                self.Inp.CrossSection["No"] == num_cs, "Iw"
            ].iloc[0]
            I_T_i = self.Inp.CrossSection.loc[
                self.Inp.CrossSection["No"] == num_cs, "It"
            ].iloc[0]

            print("Material NUmber")
            print(mat_num_i)

            E_i = self.Inp.Material.loc[self.Inp.Material["No"] == mat_num_i, "E"].iloc[
                0
            ]
            G_i = self.Inp.Material.loc[self.Inp.Material["No"] == mat_num_i, "G"].iloc[
                0
            ]
            print("Material E_i")
            print(E_i)

            K_el_i = self.ElemStem.insert_elements(
                S=0,
                E=E_i,
                G=G_i,
                A=A_i,
                I_y=I_y_i,
                I_z=I_z_i,
                I_omega=I_w_i,
                I_T=I_T_i,
                cv=0,
                z1=0,
                cw=0,
                z2=0,
                c_thet=0,
                l=length,
            )

            K_el_i = np.matmul(
                self.TransMats[i], np.matmul(K_el_i, self.TransMats[i].T)
            )

            self.K_el_i_store[i] = K_el_i

            K_11 = K_el_i[0:7, 0:7]
            K_12 = K_el_i[0:7, 7:14]
            K_21 = K_el_i[7:14, 0:7]
            K_22 = K_el_i[7:14, 7:14]

            Kp[
                7 * (node_i - 1) : 7 * (node_i - 1) + 7,
                7 * (node_i - 1) : 7 * (node_i - 1) + 7,
            ] += K_11

            Kp[
                7 * (node_i - 1) : 7 * (node_i - 1) + 7,
                7 * (node_j - 1) : 7 * (node_j - 1) + 7,
            ] += K_12

            Kp[
                7 * (node_j - 1) : 7 * (node_j - 1) + 7,
                7 * (node_i - 1) : 7 * (node_i - 1) + 7,
            ] += K_21

            Kp[
                7 * (node_j - 1) : 7 * (node_j - 1) + 7,
                7 * (node_j - 1) : 7 * (node_j - 1) + 7,
            ] += K_22

        return Kp

    def RestraintData(self):
        """
        This functions implements the restraint data, which is loaded from the \n
        input file. \n
        There are 7 DOF's per node. \n
        Therefore the restrained DOF in the global stiffness matrix can be expressed by: \n
        GDOF = 7 * (node-1) + DOF \n
        """
        res_nodes = self.Inp.RestraintData["Node"]
        res_dof = self.Inp.RestraintData["Dof"]
        res_stif = self.Inp.RestraintData["Cp[MN/m]/[MNm/m]"]

        for i in range(len(res_dof)):
            glob_dof = 7 * (res_nodes[i] - 1) + res_dof[i]
            print("restrain ", glob_dof)
            self.GesMat[glob_dof, glob_dof] += res_stif[i]

    def LocalLoadVectorLine(self):
        """
        Lokale Festendkräfte aus Linienlasten (lokales System!).
        Speichert in self.S_loc_elem_line und gibt dieses Array zurück.
        """
        n_elem = len(self.Inp.members["na"])
        self.S_loc_elem_line = np.zeros((14, n_elem), dtype=float)

        res_mbr    = self.Inp.ElementLoads["Member"]
        res_line_a = self.Inp.ElementLoads["qza"]
        res_line_b = self.Inp.ElementLoads["qze"]  # aktuell nicht benutzt

        for k in range(len(res_mbr)):
            e = int(res_mbr[k] - 1)

            L = float(self.member_length[e])
            q = float(res_line_a[k])  # TODO: falls qza != qze -> konsistente Formeln

            # Lokale Festendkräfte (deine Konvention beibehalten)
            # Vz
            self.S_loc_elem_line[3,  e]  += +q * L / 2.0
            self.S_loc_elem_line[10, e]  += +q * L / 2.0

            # My
            self.S_loc_elem_line[4,  e]  += -q * L**2 / 12.0
            self.S_loc_elem_line[11, e]  += +q * L**2 / 12.0

        return self.S_loc_elem_line
    
    def LocalLoadVectorTemp(self):
        """
        Lokale Festendkräfte aus Temperatur (lokales System!).
        Speichert in self.S_loc_elem_temp und gibt dieses Array zurück.
        """
        n_elem = len(self.Inp.members["na"])
        self.S_loc_elem_temp = np.zeros((14, n_elem), dtype=float)

        res_mbr     = self.Inp.TemperatureForces["Member"]
        res_tem_dT  = self.Inp.TemperatureForces["dT[K]"]
        res_tem_dTz = self.Inp.TemperatureForces["dTz[K]"]
        res_tem_dTy = self.Inp.TemperatureForces["dTy[K]"]

        for k in range(len(res_mbr)):
            e = int(res_mbr[k] - 1)

            EA = self.ElemStem.E * self.ElemStem.A
            EIz = self.ElemStem.E * self.ElemStem.I_z
            EIy = self.ElemStem.E * self.ElemStem.I_y

            dT  = float(res_tem_dT[k])
            dTz = float(res_tem_dTz[k])
            dTy = float(res_tem_dTy[k])

            # N
            self.S_loc_elem_temp[0, e]  += -EA  * 1e-5 * dT
            self.S_loc_elem_temp[7, e]  += +EA  * 1e-5 * dT

            # Mz
            self.S_loc_elem_temp[2, e]  += -EIz * 1e-5 * dTy
            self.S_loc_elem_temp[9, e]  += +EIz * 1e-5 * dTy

            # My
            self.S_loc_elem_temp[4, e]  += -EIy * 1e-5 * dTz
            self.S_loc_elem_temp[11,e]  += +EIy * 1e-5 * dTz

        return self.S_loc_elem_temp

    def GlobalLoadVector(self):
        """
        Globaler Lastvektor:
        - Element-Festendkräfte (Temp + Linienlast) lokal -> global via T
        - Knotenlasten (global) addieren
        """
        # erzeugt/updated self.S_loc_elem_temp / self.S_loc_elem_line
        self.LocalLoadVectorTemp()
        self.LocalLoadVectorLine()

        n_elem = len(self.Inp.members["na"])
        F_glob = np.zeros(self.Inp.nDoF, dtype=float)

        na_memb = self.Inp.members["na"]
        ne_memb = self.Inp.members["ne"]

        for e in range(n_elem):
            node_i = int(na_memb[e])
            node_j = int(ne_memb[e])

            base_i = 7 * (node_i - 1)
            base_j = 7 * (node_j - 1)

            F_e_loc = (self.S_loc_elem_temp[:, e] + self.S_loc_elem_line[:, e]).reshape(14, 1)

            T = self.TransMats[e]              # local -> global
            F_e_glob = (T @ F_e_loc).ravel()   # (14,)

            F_glob[base_i:base_i+7] += F_e_glob[0:7]
            F_glob[base_j:base_j+7] += F_e_glob[7:14]

        # Knotenlasten (global)
        res_nodes = self.Inp.NodalForces["Node"]
        res_dof   = self.Inp.NodalForces["Dof"]
        res_forc  = self.Inp.NodalForces["Value[MN/MNm]"]

        for k in range(len(res_nodes)):
            node = int(res_nodes[k])
            base = 7 * (node - 1)
            dof  = str(res_dof[k]).strip().lower()

            if dof == "fx":
                idx = base + 0
            elif dof == "fy":
                idx = base + 1
            elif dof == "fz":
                idx = base + 3
            else:
                continue

            F_glob[idx] += float(res_forc[k])

        return F_glob

    def SolveDisplacement(self):
        u_glob = np.linalg.solve(self.GesMat, self.FGes)
        return u_glob

    def StoreLocalDisplacements(self):
        u_el = np.zeros(
            [14, len(self.Inp.members["na"])]
        )  # Initialise the primary stiffness matrix

        na_memb = self.Inp.members["na"]
        ne_memb = self.Inp.members["ne"]

        for i in range(0, len(self.Inp.members["na"]), 1):
            numa = 7 * (na_memb[i] - 1)
            nume = 7 * (ne_memb[i] - 1)

            u_el[0:7, i] = self.u_ges[numa : numa + 7]
            u_el[7:14, i] = self.u_ges[nume : nume + 7]

            # u_el[:,i] = np.matmul(self.TransMats[i],u_el[:,i])

        return u_el

    def CalculateLocalInnerForces(self):
        """
        Berechnet lokale Schnittgrößen an linkem und rechtem Schnittufer:

        - f_glob = K_el_global @ u_el_global  (wie bei dir gespeichert)
        - f_loc  = T.T @ f_glob               (global -> local, da T local->global)
        - f_eff  = f_loc - (f0_temp_loc + f0_line_loc)

        Konvention:
        - Linkes Schnittufer:   s_L = - f_I,loc
        - Rechtes Schnittufer:  s_R = + f_J,loc
        - Zug positiv

        Ergebnis in self.*_el_i_store[i] jeweils als (2,1): [links; rechts]
        """

        n_elem = len(self.Inp.members["na"])
        s_el = np.zeros((14, n_elem), dtype=float)

        # Hole (lokale) Festendkräfte. Falls du die Arrays noch nicht als Attribute speicherst,
        # werden sie hier neu erzeugt.
        F_loc_temp = self.LocalLoadVectorTemp()
        F_loc_line = self.LocalLoadVectorLine()

        for e in range(n_elem):

            # 1) Element-Endkräfte aus K*u (im Koordinatensystem von K_el_i_store & u_el)
            f_glob = (self.K_el_i_store[e] @ self.u_el[:, e]).reshape(14, 1)
            s_el[:, e] = f_glob.ravel()

            # 2) global -> local
            T = self.TransMats[e]          # local -> global
            f_loc = T.T @ f_glob           # global -> local

            # 3) lokale Festendkräfte (Temp + Linienlast)
            f0_loc = (F_loc_temp[:, e] + F_loc_line[:, e]).reshape(14, 1)

            # 4) wirksame lokale Endkräfte
            f_eff = f_loc - f0_loc

            # 5) Schnittufer-Abbildung: links = -f_I, rechts = +f_J
            # Indizes laut Layout:
            # [0] Na  [1] Vya [2] Mza [3] Vza [4] Mya [5] Mxa [6] Mwa
            # [7] Nb  [8] Vyb [9] Mzb [10]Vzb [11]Myb [12]Mxb [13]Mwb

            N_L,  N_R  = -f_eff[0, 0],   f_eff[7, 0]
            Vy_L, Vy_R = -f_eff[1, 0],   f_eff[8, 0]
            Mz_L, Mz_R = -f_eff[2, 0],   f_eff[9, 0]
            Vz_L, Vz_R = -f_eff[3, 0],   f_eff[10,0]
            My_L, My_R = -f_eff[4, 0],   f_eff[11,0]
            Mx_L, Mx_R = -f_eff[5, 0],   f_eff[12,0]

            self.N_el_i_store[e]  = np.array([N_L,  N_R ]).reshape(2, 1)
            self.VY_el_i_store[e] = np.array([Vy_L, Vy_R]).reshape(2, 1)
            self.VZ_el_i_store[e] = np.array([Vz_L, Vz_R]).reshape(2, 1)

            self.MZ_el_i_store[e] = np.array([Mz_L, Mz_R]).reshape(2, 1)
            self.MY_el_i_store[e] = np.array([My_L, My_R]).reshape(2, 1)
            self.MX_el_i_store[e] = np.array([Mx_L, Mx_R]).reshape(2, 1)

        return s_el

    def MainConvergence(self):

        self.TransMats = self.CalculateTransMat()
        self.GesMat = self.BuildStructureStiffnessMatrix()

        self.RestraintData()

        # <<< HIER: Singularitäts-Check >>>
        self.check_singularity(self.GesMat, verbose=True)

        self.FGes = self.GlobalLoadVector()

        self.u_ges = self.SolveDisplacement()

        self.u_el = self.StoreLocalDisplacements()

        self.s_el = self.CalculateLocalInnerForces()

    def run(self) -> AnalysisResults:
        """
        Führt die komplette Berechnung aus und gibt ein Results-Objekt zurück.
        """
        self.MainConvergence()

        return AnalysisResults(
            Inp=self.Inp,
            u_ges=self.u_ges,
            GesMat=self.GesMat,
            FGes=self.FGes,
            TransMats=self.TransMats,
            K_el_i_store=self.K_el_i_store,
            u_el=self.u_el,
            s_el=self.s_el,
            N_el_i_store=self.N_el_i_store,
            VY_el_i_store=self.VY_el_i_store,
            VZ_el_i_store=self.VZ_el_i_store,
            MX_el_i_store=self.MX_el_i_store,
            MY_el_i_store=self.MY_el_i_store,
            MZ_el_i_store=self.MZ_el_i_store,
            member_length=np.array(self.member_length, dtype=float),
        )

    def check_global_equilibrium(self):
        r = (self.GesMat @ self.u_ges - self.FGes).ravel()  # Reaktionsvektor auf allen DOFs

        # welche DOFs sind "gelagert"? (klassisch: Cp groß oder Fix)
        # Bei dir: RestraintData addiert Cp auf Diagonale.
        # -> als Näherung: DOF ist gelagert, wenn Cp > 0 in Input.
        restrained = np.zeros_like(r, dtype=bool)
        for _, row in self.Inp.RestraintData.iterrows():
            gdof = 7*(int(row["Node"])-1) + int(row["Dof"])
            restrained[gdof] = True

        free = ~restrained

        print("max residual on FREE dofs:", np.max(np.abs(r[free])))
        print("sum residual on FREE dofs:", np.sum(r[free]))

    def debug_single_fx(self, node=2, Fx=1.0):
        # setze alle Lasten leer
        self.Inp.ElementLoads = self.Inp.ElementLoads.iloc[0:0].copy()
        self.Inp.TemperatureForces = self.Inp.TemperatureForces.iloc[0:0].copy()
        self.Inp.NodalForces = self.Inp.NodalForces.iloc[0:0].copy()

        # eine Last
        import pandas as pd
        self.Inp.NodalForces = pd.DataFrame([{"Node":node, "Dof":"Fx", "Value[MN/MNm]":Fx}])

        self.MainConvergence()

        ux = self.u_ges[7*(node-1) + 0]
        uy = self.u_ges[7*(node-1) + 1]
        uz = self.u_ges[7*(node-1) + 3]
        print("ux, uy, uz =", ux, uy, uz)

    def sum_reactions_fx(self):
        r = (self.GesMat @ self.u_ges - self.FGes).ravel()

        Rx = 0.0
        for _, row in self.Inp.RestraintData.iterrows():
            node = int(row["Node"])
            dof  = int(row["Dof"])
            if dof == 0:  # ux-DOF
                Rx += r[7*(node-1) + 0]

        Fx = 0.0
        for _, row in self.Inp.NodalForces.iterrows():
            if str(row["Dof"]).strip().lower() == "fx":
                Fx += float(row["Value[MN/MNm]"])

        print("Sum Fx (applied) =", Fx)
        print("Sum Rx (support) =", Rx)
        print("Fx + Rx          =", Fx + Rx)

    def sum_spring_reactions_fx(self):
        u = self.u_ges.ravel()

        Rx = 0.0
        Fx = 0.0

        # aufgebrachte nodale Fx
        for _, row in self.Inp.NodalForces.iterrows():
            if str(row["Dof"]).strip().lower() == "fx":
                Fx += float(row["Value[MN/MNm]"])

        # Federkräfte aus RestraintData: k * u an genau diesen DOFs
        for _, row in self.Inp.RestraintData.iterrows():
            node = int(row["Node"])
            dof  = int(row["Dof"])
            k    = float(row["Cp[MN/m]/[MNm/m]"])
            gdof = 7*(node-1) + dof

            if dof == 0:  # ux
                Rx += k * u[gdof]

        print("Sum Fx (applied) =", Fx)
        print("Sum spring Rx    =", Rx)
        print("Fx - Rx          =", Fx - Rx)






import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy.interpolate import interp1d
from Cross_Section_Kappa import M_Kappa_Cross_Section
from Querschnittsbreite import PolygonSection

class Druckgliedbemessung_KGV:
    def __init__(
        self,
        _LoadingType="csv_druckglied",
        Iter="Bruch",
        Reinforcement="csv",
        Querschnittsbemessung="Polygon",
    ):
        self.Bewehrung = Reinforcement
        self.LoadingType = _LoadingType

        self.Querschnittswerte(0.5)
        self.Bewehrungsangaben()
        self.ReadLoading()
        self.Calculate_M_Kappa_Interpolation()
        self.Calculate_Cross_Section_Capacity()
        self.Column_Parameters()
        self.Iteration_Second_Order()

    def Querschnittswerte(self, _height_test=0.5):
        # Example usage
        df = pd.read_csv("Polygon/Polygon.csv")
        self.vertices = df.values.tolist()

        self.polygon = PolygonSection(self.vertices)

        # Calculate the height of the polygon based on y-values

        y_values = [vertex[1] for vertex in self.vertices]
        self.height = abs(max(y_values) - min(y_values))
        self.z_cu = self.height - self.polygon.centroid[1]
        self.z_co = self.height - self.z_cu

        # Define the rotation angle
        self.angle_degrees = 0
        self.polygon.rotate(self.angle_degrees)

        height = _height_test
        section_width = self.polygon.calculate_section_width_at_height(height / 2)

        self.h_c = height
        self.b_c = section_width

        self.A = self.polygon.A

        return section_width

    def Bewehrungsangaben(self):
        if self.Bewehrung == "csv":
            df = pd.read_csv("Bewehrung/Linienbewehrung.csv")
            print(df)
            for i in range(0, len(df["Lage"]), 1):
                Lage = df["Lage"][i]
                if Lage == "Unten":
                    self.d_s1 = (
                        df["dsl [m]"][i] * 0.5 + df["cnom [m]"][i] + df["dsw [m]"][i]
                    )
                    self.A_s1 = df["As [cm**2]"][i] * 0.01**2
                elif Lage == "Oben":
                    self.d_s2 = (
                        df["dsl [m]"][i] * 0.5 + df["cnom [m]"][i] + df["dsw [m]"][i]
                    )
                    self.A_s2 = df["As [cm**2]"][i] * 0.01**2

            self.z_ds1 = self.z_cu - self.d_s1
            self.z_ds2 = self.z_co - self.d_s2

    def ReadLoading(self):
        if self.LoadingType == "csv_druckglied":
            df = pd.read_csv("Lasteingabe/Lasten_Druckglied.csv")
            print(df)
            self.Zugseite = None
            for i in range(0, len(df["Grenzzustand"]), 1):
                if df["Grenzzustand"][i] == "GZT":
                    self.NEd_GZT = df["NEd in [MN]"][i]
                    self.MEd1_GZT = df["MEd1 in [MNm]"][i]
                    self.MEd2_GZT = df["MEd2 in [MNm]"][i]
                elif df["Grenzzustand"][i] == "GZG":
                    self.NEd_GZG = df["NEd in [MN]"][i]
                    self.MEd1_GZG = df["MEd1 in [MNm]"][i]
                    self.MEd2_GZG = df["MEd2 in [MNm]"][i]

        else:
            self.NEd_GZT = float(
                input("Geben Sie die Normalkraft NEd im GZT in [MN] ein: \n")
            )
            self.MEd_GZT1 = float(
                input("Geben Sie das Biegemoment an Stelle 1 im GZT in [MN] ein: \n")
            )

            self.MEd_GZT2 = float(
                input("Geben Sie das Biegemoment an Stelle 2 im GZT in [MN] ein: \n")
            )

    def Calculate_M_Kappa_Interpolation(self):
        self.M_Kappa = M_Kappa_Cross_Section("csv_druckglied")

    def Calculate_Cross_Section_Capacity(self):
        try:
            self.M_Rd = self.M_Kappa.M_D
            print("The maximum Moment for the cross section is: ", self.M_Rd)
        except:
            self.M_Rd = self.M_Kappa.M_C
            print("The maximum Moment for the cross section is: ", self.M_Rd)

    def Column_Parameters(self):
        df = pd.read_csv("Systemgeometrie/Druckglied.csv")
        self.l_col = df["lcol in [m]"].iloc[0]

    def Iteration_Second_Order(self):
        M_Ed1_0 = self.MEd1_GZT
        M_Ed1_1 = self.MEd1_GZT
        M_Ed2_0 = self.MEd2_GZT

        self.M_0_List = np.linspace(M_Ed1_0, M_Ed2_0, 100)
        self.M_1_List = np.linspace(self.l_col, 0, 100)

        kappa_1 = float(self.M_Kappa.Interpolation_Function(M_Ed1_1)[0])
        kappa_2 = float(self.M_Kappa.Interpolation_Function(M_Ed2_0)[0])

        print("KAPPA 1", kappa_1)

        x_list = np.linspace(0, float(self.l_col), 100)

        # kappa_list = [kappa_1 - (kappa_1 - kappa_2) / 99 * i for i in range(0, 100, 1)]
        kappa_list = [kappa_1 - (kappa_1 - kappa_2) / 99 * i for i in range(100)]


        d_0 = np.trapz(self.M_1_List * kappa_list, x_list)

        iter = 0
        while iter < 100:
            print(M_Ed1_1 - M_Ed1_0)
            M_Ed1_0 = self.MEd1_GZT

            d_1 = np.trapz(self.M_1_List * kappa_list, x_list)

            M_Ed1_1 = self.MEd1_GZT + d_1 * abs(self.NEd_GZT)

            print("M0", M_Ed1_0)
            print("M1", M_Ed1_1)

            self.M_0_List = np.linspace(M_Ed1_1, M_Ed2_0, 100)

            for i in range(0, 100, 1):
                kappa_i = float(self.M_Kappa.Interpolation_Function(
                    M_Ed1_1 - (M_Ed1_1 - M_Ed2_0) / 99 * i
                )[0])
                kappa_list[i] = kappa_i

            if abs(M_Ed1_1 - M_Ed1_0) < 1e-3:
                break

            iter += 1

        print("Iteration ", iter)
        print("Deflection ", d_1)

        print("Moment ", M_Ed1_1)

        plt.plot(kappa_list)
        plt.show()

        kappa_value, M_value, kappa,kappa_new,M_interpolated,M =self.M_Kappa.Interpolation_Function(M_Ed1_1)
        self.M_Kappa.Plot_Interpolation_M_Kappa(kappa_value, M_value, kappa,kappa_new,M_interpolated,M)



Bemess = Druckgliedbemessung_KGV()

print(Bemess.M_Kappa.Baustoff.Ecm)
print(Bemess.M_Kappa.Baustoff.fctm)


from Programm import mainloop
from Plotting import StructurePlotter
from Output_Data import OutputData
import matplotlib
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt


if __name__ == "__main__":
    calc = mainloop()
    res = calc.run()

    plotter = StructurePlotter(res)
    Output = OutputData()

    calc.check_global_equilibrium()

    # print(type(res), hasattr(res, "Inp"), hasattr(res, "GesMat"), hasattr(res, "u_ges"), hasattr(res, "FGes"))

    # res = calc.run()              

    out = OutputData()
    df = out.support_reactions_from_springs_table(res)
    print(df)

    calc.sum_reactions_fx()
    calc.sum_spring_reactions_fx()

    plotter.plot_support_reactions_2d_interactive()
    plt.show()

    plotter.plot_nodal_loads_2d_interactive()

    plt.show()

    fig, ax, s = plotter.plot_endforces_2d_interactive(kind="MY", scale_init=0.2, node_labels=True, elem_labels=True)

    plt.show()

    fig, ax, s = plotter.plot_endforces_2d_interactive(kind="N", scale_init=0.2, node_labels=True, elem_labels=True)

    plt.show()

    fig, ax, s = plotter.plot_endforces_2d_interactive(kind="VZ", scale_init=0.2, node_labels=True, elem_labels=True)

    plt.show()










# DEPENDENCIES
import copy  # Allows us to create copies of objects in memory
import math  # Math functionality
import numpy as np  # Numpy for working with arrays
import matplotlib.pyplot as plt  # Plotting functionality
import matplotlib.colors  # For colormap functionality
import ipywidgets as widgets
from glob import glob  # Allows check that file exists before import
from numpy import genfromtxt  # For importing structure data from csv
import pandas as pd

class Input:
    def __init__(self):
        print("Input-Class")
        self.NodalData()
        self.EdgeData()
        self.RestraintData()
        self.NodalForces()
        self.TemperatureLoads()
        self.ElementLoads()
        self.MaterialData()
        self.CrossSectionData()

    def MaterialData(self):
        try:
            df = pd.read_csv("data/Material.csv")
        except:
            df = pd.DataFrame([0])

        self.Material = df

    def CrossSectionData(self):
        try:
            df = pd.read_csv("data/CrossSection.csv")
        except:
            df = pd.DataFrame([0])

        self.CrossSection = df

    def NodalData(self):
        if glob("data/Vertices.csv"):
            # Annahme: Die erste Zeile enthält Header
            df = pd.read_csv("data/Vertices.csv")

            # Wähle nur die numerischen Spalten aus (z.B. 'Node' und 'P[N]')
            numeric_columns = ['x[m]', 'y[m]','z[m]']
            df_numeric = df[numeric_columns]

            self.nodes = df_numeric
            self.nDoF = (len(self.nodes))*7
            print("1. 🟢 Vertices.csv imported")
        else:
            self.forceLocationData = []
            print("1. ⚠️ Vertices.csv not found")

    def EdgeData(self):
        # MANDATORY IMPORT: member definitions
        if glob("data/Edges.csv"):
            df = pd.read_csv("data/Edges.csv")
            numeric_columns = ['na', 'ne',"cs"]
            df_numeric = df[numeric_columns]
            self.members = df_numeric
            
            print("2. 🟢 Edges.csv imported")
        else:
            print("2. 🛑 STOP: Edges.csv not found")

    def RestraintData(self):
        df = pd.read_csv("data/Restraint_Data.csv")
        # Wähle nur die numerischen Spalten aus (z.B. 'Node' und 'P[N]')
        numeric_columns = ['Node', 'Dof','Cp[MN/m]/[MNm/m]']
        df_numeric = df[numeric_columns]
        self.RestraintData = df_numeric

    def NodalForces(self):
        df = pd.read_csv("data/NodalForces.csv")
        self.NodalForces = df

    def ElementLoads(self):
        try:
            df = pd.read_csv("data/Linienlasten.csv")
        except:
            df = pd.DataFrame([0])

        self.ElementLoads = df

    def TemperatureLoads(self):
        try:
            df = pd.read_csv("data/TemperatureLoading.csv")
        except:
            df = pd.DataFrame([0])
        self.TemperatureForces = df



import pandas as pd
import numpy as np

class OutputData:
    def __init__(self):
        print("Init OutputData")

    def support_reactions_from_springs_table(self, res, u0_default=0.0):
        # --- robuster Typ-/Attribut-Check ---
        needed = ("Inp", "GesMat", "u_ges", "FGes")
        missing = [a for a in needed if not hasattr(res, a)]
        if missing:
            raise TypeError(
                "Du musst ein AnalysisResults-Objekt übergeben (res = mainloop().run()). "
                f"Übergeben wurde: {type(res)}; fehlende Attribute: {missing}"
            )

        dfR = res.Inp.RestraintData.copy()

        # u als 1D
        u = np.asarray(res.u_ges, dtype=float).reshape(-1)

        # DOF Mapping (7 dof/node)
        dof_name = {
            0: ("Fx", "MN"),
            1: ("Fy", "MN"),
            2: ("Mz", "MNm"),
            3: ("Fz", "MN"),
            4: ("My", "MNm"),
            5: ("Mx", "MNm"),
            6: ("W",  "-"),
        }

        rows = []
        for _, row in dfR.iterrows():
            node = int(row["Node"])
            dof  = int(row["Dof"])
            k    = float(row["Cp[MN/m]/[MNm/m]"])

            gdof = 7 * (node - 1) + dof
            ui   = float(u[gdof])
            u0   = float(u0_default)

            Ri = k * (ui - u0)

            name, unit = dof_name.get(dof, (f"DOF{dof}", ""))

            rows.append({
                "Node": node,
                "DOF": dof,
                "Type": name,
                "k": k,
                "u": ui,
                "R_spring": Ri,
                "Unit": unit
            })

        return pd.DataFrame(rows).sort_values(["Node", "DOF"]).reset_index(drop=True)



import sympy as sp
import numpy as np
import math

class ElemStema:
    def __init__(self):

        # Initialisierung der Steifigkeitsmatrix
        self.Ke = np.zeros((14, 14))

    def TransformationMatrix(self, posI, posJ):
        Pi = np.asarray(posI, dtype=float).reshape(3,)
        Pj = np.asarray(posJ, dtype=float).reshape(3,)

        v = Pj - Pi
        L = np.linalg.norm(v)

        T = np.zeros((14, 14), dtype=float)

        # identische Knoten
        if L < 1e-12:
            np.fill_diagonal(T, 1.0)
            return T

        # lokale x-Achse
        ex = v / L

        # robuste Referenzachse wählen (nicht parallel zu ex)
        g = np.array([0.0, 1.0, 0.0], dtype=float)  # bevorzugt global y
        if abs(np.dot(ex, g)) > 0.95:
            g = np.array([0.0, 0.0, 1.0], dtype=float)  # sonst global z
            if abs(np.dot(ex, g)) > 0.95:
                g = np.array([1.0, 0.0, 0.0], dtype=float)  # zur Not global x

        # --- Stabiler Aufbau: erst ey als Projektion von g orthogonal zu ex ---
        ey = g - np.dot(g, ex) * ex
        ney = np.linalg.norm(ey)
        if ney < 1e-12:
            raise ValueError("Degeneriertes lokales Achsensystem (ey ~ 0).")
        ey = ey / ney

        # --- Rechtshändiges System erzwingen ---
        ez = np.cross(ex, ey)
        nez = np.linalg.norm(ez)
        if nez < 1e-12:
            raise ValueError("Degeneriertes lokales Achsensystem (ez ~ 0).")
        ez = ez / nez

        # Optional: Flip-Schutz (ey soll möglichst in Richtung g zeigen)
        # verhindert, dass bei ähnlichen Elementen ey/ez um 180° springen
        if np.dot(ey, g) < 0.0:
            ey *= -1.0
            ez *= -1.0

        # R: local -> global (Spalten sind ex, ey, ez)
        R = np.column_stack((ex, ey, ez))

        # ---- 6x6 Transformationsblock passend zu deiner DOF-Reihenfolge ----
        Ti = np.zeros((6, 6), dtype=float)

        # translations: [ux, uy, uz] sitzen auf [0,1,3]
        trans = [0, 1, 3]
        # rotations: [rx, ry, rz] sitzen auf [5,4,2]
        rot = [5, 4, 2]

        Ti[np.ix_(trans, trans)] = R
        Ti[np.ix_(rot,  rot )]  = R

        # In 14x14 einbauen (Knoten a: 0..6, Knoten b: 7..13)
        T[0:6, 0:6]   = Ti
        T[7:13, 7:13] = Ti

        # Warping
        T[6, 6] = 1.0
        T[13, 13] = 1.0

        # Debug-Sanity
        if not np.isfinite(T).all():
            print("TM NaN/Inf:", posI, posJ)
            print(T)
            raise ValueError("TM enthält NaN/Inf")

        return T

    
    
    def insert_elements(self, S, E, G, A, I_y, I_z, I_omega, I_T, cv, z1, cw, z2, c_thet, l):
        """
        Element-Stiffness-Matrix:
        Na
        Vya
        Mza
        Vza
        Mya
        Mxa
        Mwa
        Nb
        Vyb
        Mzb
        Vzb
        Myb
        Mxb
        Mwb
        """
        self.Ke[:, :] = 0.0
        self.S = S  # Stiffness of shear field

        self.E = E  # Material stiffness of the beam
        self.G = G

        self.A = A
        self.I_y = I_y
        self.I_z = I_z

        self.I_omega = I_omega
        self.I_T = I_T

        self.cv = cv
        self.z1 = z1
        self.cw = cw
        self.z2 = z2

        self.c_thet = c_thet

        self.l = l
        # Matrixeinträge gemäß Tabelle definieren
        self.Ke[0, 0] = self.Ke[7, 7] = self.E * self.A / self.l
        self.Ke[0, 7] = self.Ke[7, 0] = -self.E * self.A / self.l

        self.Ke[1, 1] = self.Ke[8, 8] = (
            12 * self.E * self.I_z / self.l**3 + 13 / 35 * self.cv * self.l + 1.2 * self.S / self.l
        )
        self.Ke[1, 2] = 6 * self.E * self.I_z / self.l**2   + 11 / 210 * self.cv * self.l**2 + 0.1 * self.S

        self.Ke[1, 5] = 13 / 35 * self.cv * self.l * self.z1 - 1.2 * self.S / self.l * self.z2
        self.Ke[1, 6] = -11 / 210 * self.cv * self.l**2 * self.z1 + 0.1 * self.S * self.z2

        self.Ke[1, 8] = -12 * self.E * self.I_z / self.l**3 + 9 / 70 * self.cv * self.l - 1.2 * self.S / self.l
        self.Ke[1, 9] = 6 * self.E * self.I_z / self.l**2 - 13 / 420 * self.cv * self.l**2 + 0.1 * self.S

        self.Ke[1, 12] = 9 / 70 * self.cv * self.l * self.z1 + 1.2 * self.S / self.l * self.z2
        self.Ke[1, 13] = 13 / 420 * self.cv * self.l**2 * z1 + 0.1 * self.S * self.z2

        self.Ke[2, 2] = 4 * self.E * self.I_z / self.l + 1 / 105 * self.cv * self.l**3 + 2 / 15 * self.S * self.l
        self.Ke[9, 9] = self.Ke[2, 2]

        self.Ke[2, 5] = 11 / 210 * self.cv * l**2 * self.z1 - 0.1 * self.S * self.z2
        self.Ke[2, 6] = -1 / 105 * self.cv * self.l**3 * self.z1 + 2 / 15 * self.S * self.l * self.z2
        self.Ke[2, 8] = -6 * self.E * self.I_z / self.l**2 + 13 / 420 * self.cv * l**2 - 0.1 * self.S

        self.Ke[2, 9] = 2 * self.E * self.I_z / self.l - 1 / 140 * self.cv * self.l**3 - 1 / 30 * self.S * self.l
        self.Ke[2, 12] = 13 / 420 * self.cv * self.l**2 * self.z1 + 0.1 * self.S * self.z2
        self.Ke[2, 13] = 1 / 140 * self.cv * self.l**3 * self.z1 - 1 / 30 * self.S * self.l * self.z2

        self.Ke[3, 3] = 12 * self.E * self.I_y / self.l**3 + 13 / 35 * self.cw * self.l
        self.Ke[10, 10] = self.Ke[3, 3]

        self.Ke[3, 4] = -6 * self.E * self.I_y / self.l**2 - 11 / 210 * self.cw * self.l**2
        self.Ke[3, 10] = -12 * self.E * self.I_y / self.l**3 + 9 / 70 * self.cw * self.l
        self.Ke[3, 11] = -6 * self.E * self.I_y / self.l**2 + 13 / 420 * self.cw * self.l**2

        self.Ke[4, 4] = 4 * self.E * self.I_y / self.l + 1 / 105 * self.cw * self.l**3
        self.Ke[11, 11] = self.Ke[4, 4]

        self.Ke[4, 10] = 6 * self.E * self.I_y / self.l**2 - 13 / 420 * self.cw * self.l**2
        self.Ke[4, 11] = 2 * self.E * self.I_y / self.l - 1 / 140 * self.cw * self.l**3

        self.Ke[5, 5] = self.Ke[12, 12] = (
            12 * self.E * self.I_omega / self.l**3
            + 1.2 * self.G * self.I_T / self.l
            + 13 / 35 * self.c_thet * self.l
            + 13 / 35 * self.cv * self.l * self.z1**2
            + 1.2 * self.S / self.l * self.z2**2
        )
        self.Ke[5, 6] = (
            -6 * self.E * self.I_omega / self.l**2
            - 0.1 * self.G * self.I_T
            - 11 / 210 * self.c_thet * self.l**2
            - 11 / 210 * self.cv * self.l**2 * self.z1**2
            - 0.1 * self.S * self.z2**2
        )
        self.Ke[5, 8] = 9 / 70 * self.cv * l * self.z1 + 1.2 * self.S / self.l * self.z2
        self.Ke[5, 9] = -13 / 420 * self.cv * self.l * self.z1 - 0.1 * self.S * self.z2
        self.Ke[5, 12] = (
            -12 * self.E * self.I_omega / self.l**3
            - 1.2 * self.G * self.I_T / self.l
            + 9 / 70 * self.c_thet * self.l
            + 9 / 70 * self.cv * self.l * self.z1**2
            - 1.2 * self.S / self.l * self.z2**2
        )
        self.Ke[5, 13] = (
            -6 * self.E * self.I_omega / self.l**2
            - 0.1 * self.G * self.I_T
            + 13 / 420 * self.c_thet * self.l**2
            + 13 / 420 * self.cv * self.l**2 * self.z1**2
            - 0.1 * self.S * self.z2**2
        )
        self.Ke[6, 6] = self.Ke[13, 13] = (
            4 * self.E * self.I_omega / self.l
            + 2 / 15 * self.G * self.I_T * self.l
            + 1 / 105 * self.c_thet * self.l**3
            + 1 / 105 * self.cv * self.l**3 * self.z1
            + 2 / 15 * self.S * self.l * self.z2**2
        )

        self.Ke[6, 8] = -13 / 420 * self.cv * self.l**2 * self.z1 - 0.1 * self.S * self.z2
        self.Ke[6, 9] = 1 / 140 * self.cv * self.l**3 * self.z1 - 1 / 30 * self.S * self.l * self.z2
        self.Ke[6, 12] = (
            6 * self.E * self.I_omega / self.l**2
            + 0.1 * self.G * self.I_T
            - 13 / 420 * self.c_thet * self.l**2
            - 13 / 420 * self.cv * self.l**2 * self.z1**2
            + 0.1 * self.S * self.z2**2
        )
        self.Ke[6, 13] = (
            2 * self.E * self.I_omega / self.l
            - 1 / 30 * self.G * self.I_T * self.l
            - 1 / 140 * self.c_thet * self.l**3
            - 1 / 140 * self.cv * l**3 * self.z1**2
            - 1 / 30 * self.S * self.l * self.z2**2
        )
        self.Ke[8, 9] = -6 * self.E * self.I_z / self.l**2 - 11 / 210 * self.cv * self.l**2 - 0.1 * self.S
        self.Ke[8, 12] = 13 / 35 * self.cv * self.l * self.z1 - 1.2 * self.S / self.l * self.z2
        self.Ke[8, 13] = 11 / 210 * self.cv * self.l**2 * self.z1 - 0.1 * self.S * self.z2
        self.Ke[9, 12] = -11 / 210 * self.cv * self.l**2 * self.z1 + 0.1 * self.S * self.z2
        self.Ke[9, 13] = -1 / 105 * self.cv * self.l**3 * z1 + 2 / 15 * self.S * self.l * self.z2
        self.Ke[10, 11] = 6 * E * I_y / l**2 + 11 / 210 * cw * l**2
        self.Ke[12, 13] = (
            6 * E * I_omega / l**2
            + 0.1 * G * I_T
            + 11 / 210 * c_thet * l**2
            + 11 / 210 * cv * l**2 * z1**2
            + 0.1 * S * z2**2
        )

        # Elem Matrix is symmetrical

        for i in range(14):
            for j in range(i):
                self.Ke[i, j] = self.Ke[j, i]

        return self.Ke

    def print_elem_matrix(self):
        sp.pprint(self.Ke)
